# pylint: disable=invalid-overridden-method

import tempfile
from io import BytesIO
from typing import Any

import duckdb

from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from models.models import (
    PRIMARY_EXTERNAL_ORG,
)
from models.models.participant import ParticipantUpsertInternal
from models.models.sample import SampleUpsertInternal
from test.testbase import DbIsolatedTest, run_as_sync


def query_parquet(tables: dict[str, BytesIO], query: str) -> list[dict[str, Any]]:
    """
    Handle getting data out of a parquet file, this writes the file to a temporary
    folder and then reads it back in to duckdb to get the data. This probably isn't
    the most efficient, but the files are small so it shouldn't be too slow.
    """
    duck = duckdb.connect()

    with tempfile.TemporaryDirectory() as td:
        for table_name, table_bytes in tables.items():
            filename = f'{td}/{table_name}.parquet'
            with open(filename, 'wb') as f:
                table_bytes.seek(0)
                file_val = table_bytes.getvalue()
                f.write(file_val)
                f.close()

            duck.register(table_name, duck.read_parquet(filename))

        return duck.query(query).arrow().to_pylist()


class TestMetaTable(DbIsolatedTest):
    """Test meta table operations"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()
        self.sl = SampleLayer(self.connection)
        self.pl = ParticipantLayer(self.connection)

    @run_as_sync
    async def test_export_participants(self):
        """
        Test getting a participant table from the export layer
        """

        await self.pl.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01', 'other': 'OTHER1'},
                    reported_sex=2,
                    karyotype='XX',
                    meta={'field': 1},
                ),
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX02'},
                    reported_sex=1,
                    karyotype='XY',
                    meta={'field': 2},
                ),
            ]
        )

        pts = await self.pl.export_participant_table(self.project_id)
        assert pts
        result = query_parquet(
            {'participants': pts},
            'SELECT * FROM participants order by participant_id',
        )

        self.assertEqual(2, len(result))
        self.assertEqual(1, result[0]['meta_field'])
        self.assertEqual(2, result[1]['meta_field'])
        self.assertEqual('EX01', result[0]['external_id'])
        self.assertEqual('EX02', result[1]['external_id'])
        self.assertEqual('OTHER1', result[0]['external_id_other'])

    @run_as_sync
    async def test_export_empty_table(self):
        """
        Test that exporting an empty table returns none
        """

        pts = await self.pl.export_participant_table(self.project_id)
        self.assertIsNone(pts)

    @run_as_sync
    async def test_export_samples(self):
        """
        Test getting a sample table from the export layer
        """

        await self.sl.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'field_1': 'field_1_value'},
            )
        )

        await self.sl.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test02'},
                type='blood',
                active=True,
                meta={'field_2': 'field_2_value'},
            )
        )

        samples = await self.sl.export_sample_table(self.project_id)
        assert samples
        result = query_parquet(
            {'samples': samples},
            'SELECT * FROM samples order by sample_id',
        )

        self.assertEqual(2, len(result))
        self.assertEqual('field_1_value', result[0]['meta_field_1'])
        self.assertEqual('field_2_value', result[1]['meta_field_2'])
        self.assertEqual('Test01', result[0]['external_id'])
        self.assertEqual('Test02', result[1]['external_id'])



"""Core audit analysis business logic."""

from .file_matcher import FileMatchingService
from metamist.audit.models import (
    SequencingGroup,
    Analysis,
    AuditReportEntry,
    AuditResult,
    FileMetadata,
    MovedFile,
)


class AuditAnalyzer:
    """Core audit analysis service with pure business logic."""

    def __init__(self):
        """
        Initialize the audit analyzer.
        """
        self.file_matcher = FileMatchingService()

    def analyze_sequencing_groups(
        self,
        sequencing_groups: list[SequencingGroup],
        bucket_files: list[FileMetadata],
        analyses: list[Analysis],
        excluded_sg_ids: list[str] | None = None,
    ) -> AuditResult:
        """
        Main analysis entry point for audit.

        Args:
            sequencing_groups: List of sequencing groups from Metamist
            bucket_files: List of files found in the bucket
            analyses: List of completed analyses
            excluded_sg_ids: Optional list of SG IDs to exclude

        Returns:
            AuditResult containing all findings
        """
        excluded_sg_ids = excluded_sg_ids or []

        # Update SGs with CRAM analyses
        self._update_sgs_with_cram_analyses(sequencing_groups, analyses)

        # Identify unaligned SGs
        unaligned_sgs = [sg for sg in sequencing_groups if not sg.is_complete]

        # Get all read files from Metamist
        metamist_read_files = self._get_all_read_files(sequencing_groups)

        # Find moved files
        moved_files = self.file_matcher.find_moved_files(
            metamist_read_files, bucket_files
        )

        # Find original analysis files
        self.file_matcher.find_original_analysis_files(
            analyses,
            bucket_files,
        )

        # Find uningested files
        analysis_metadata = [a.output_file for a in analyses if a.output_file]
        uningested_files = self.file_matcher.find_uningested_files(
            metamist_read_files, bucket_files, moved_files, analysis_metadata
        )

        # Generate report entries
        files_to_delete = self._generate_delete_entries(
            sequencing_groups, bucket_files, moved_files, analyses, excluded_sg_ids
        )

        files_to_review = self._generate_review_entries(
            sequencing_groups, uningested_files
        )

        moved_entries = self._generate_moved_entries(
            sequencing_groups, moved_files, excluded_sg_ids
        )

        return AuditResult(
            files_to_delete=files_to_delete,
            files_to_review=files_to_review,
            moved_files=moved_entries,
            unaligned_sequencing_groups=unaligned_sgs,
        )

    def _update_sgs_with_cram_analyses(
        self, sequencing_groups: list[SequencingGroup], analyses: list[Analysis]
    ):
        """Update sequencing groups with their CRAM analyses."""
        # Create lookup of analyses by SG ID
        cram_by_sg: dict[str, Analysis] = {}
        for analysis in analyses:
            if analysis.is_cram and analysis.sequencing_group_id:
                sg_id = analysis.sequencing_group_id
                # Keep the latest if multiple exist
                if sg_id not in cram_by_sg:
                    cram_by_sg[sg_id] = analysis

        # Update SGs
        for sg in sequencing_groups:
            if sg.id in cram_by_sg:
                sg.set_cram_analysis(cram_by_sg[sg.id])

    def _get_all_read_files(
        self, sequencing_groups: list[SequencingGroup]
    ) -> list[FileMetadata]:
        """Get all read files from all sequencing groups."""
        read_files = []
        for sg in sequencing_groups:
            read_files.extend(sg.get_all_read_files())
        return read_files

    def _generate_delete_entries(
        self,
        sequencing_groups: list[SequencingGroup],
        bucket_files: list[FileMetadata],
        moved_files: dict[str, MovedFile],
        analyses: list[Analysis],
        excluded_sg_ids: list[str],
    ) -> list[AuditReportEntry]:
        """Generate report entries for files to delete."""
        entries = []

        bucket_paths = set()
        for path in bucket_files:
            bucket_paths.add(str(path.filepath))
        for path in moved_files.values():
            bucket_paths.add(str(path.new_path))

        for sg in sequencing_groups:
            # Assay files from completed SGs can be deleted
            if sg.id in excluded_sg_ids or not sg.is_complete:
                continue
            for assay in sg.assays:
                for read_file in assay.read_files:
                    if str(read_file.filepath) not in bucket_paths:
                        continue
                    entry = self._create_report_entry(
                        sg=sg, assay_id=assay.id, file_metadata=read_file
                    )
                    entries.append(entry)

            # Original files from ingested analyses can be deleted
            for analysis in analyses:
                if not analysis.original_file or analysis.sequencing_group_id != sg.id:
                    continue
                entry = self._create_report_entry(
                    sg=sg, file_metadata=analysis.original_file
                )
                entries.append(entry)

        return entries

    def _generate_review_entries(
        self,
        sequencing_groups: list[SequencingGroup],
        uningested_files: list[FileMetadata],
    ) -> list[AuditReportEntry]:
        """Generate report entries for files to review."""
        entries = []

        # Build lookup of external IDs to completed SGs
        completed_sgs_by_sample_id: dict[str, SequencingGroup] = {}
        completed_sgs_by_participant_id: dict[str, SequencingGroup] = {}

        for sg in sequencing_groups:
            if not sg.is_complete:
                continue

            for ext_id in sg.sample.external_ids.values():
                completed_sgs_by_sample_id[ext_id] = sg

            for ext_id in sg.sample.participant.external_ids.values():
                completed_sgs_by_participant_id[ext_id] = sg

        # Check each uningested file
        for file_metadata in uningested_files:
            filename = file_metadata.filepath.name
            matching_sg = None

            # Check if filename contains a known sample ID
            for sample_id, sg in completed_sgs_by_sample_id.items():
                if sample_id in filename:
                    matching_sg = sg
                    break

            # Check if filename contains a known participant ID
            if not matching_sg:
                for participant_id, sg in completed_sgs_by_participant_id.items():
                    if participant_id in filename:
                        matching_sg = sg
                        break

            if matching_sg:
                # File matches a completed SG, should be deleted not ingested
                entry = self._create_report_entry(
                    sg=matching_sg, file_metadata=file_metadata
                )
            else:
                # File doesn't match any SG, should be ingested
                entry = AuditReportEntry(
                    filepath=str(file_metadata.filepath),
                    filesize=file_metadata.filesize,
                )

            entries.append(entry)

        return entries

    def _generate_moved_entries(
        self,
        sequencing_groups: list[SequencingGroup],
        moved_files: dict[str, MovedFile],
        excluded_sg_ids: list[str],
    ) -> list[AuditReportEntry]:
        """Generate report entries for moved files."""
        entries = []

        # Build lookup of file paths to assays and SGs
        path_to_sg_assay: dict[str, tuple[SequencingGroup, int]] = {}

        for sg in sequencing_groups:
            for assay in sg.assays:
                for read_file in assay.read_files:
                    path = str(read_file.filepath)
                    path_to_sg_assay[path] = (sg, assay.id)

        # Create entries for moved files
        for original_path, moved_file in moved_files.items():
            if original_path not in path_to_sg_assay:
                continue

            sg, assay_id = path_to_sg_assay[original_path]

            if sg.id in excluded_sg_ids or not sg.is_complete:
                continue

            entry = self._create_report_entry(
                sg=sg, assay_id=assay_id, file_metadata=moved_file.metadata
            )
            entries.append(entry)

        return entries

    def _create_report_entry(
        self,
        sg: SequencingGroup,
        file_metadata: FileMetadata,
        assay_id: int | None = None,
    ) -> AuditReportEntry:
        """Create a report entry from sequencing group and file data."""
        return AuditReportEntry(
            filepath=str(file_metadata.filepath),
            filesize=file_metadata.filesize,
            sg_id=sg.id,
            sg_type=sg.type,
            sg_tech=sg.technology,
            sg_platform=sg.platform,
            assay_id=assay_id,
            cram_analysis_id=sg.cram_analysis.id if sg.cram_analysis else None,
            cram_file_path=sg.cram_path,
            sample_id=sg.sample.id,
            sample_external_ids=sg.sample.external_ids.ids,
            participant_id=sg.sample.participant.id,
            participant_external_ids=sg.sample.participant.external_ids.ids,
        )



"""GraphQL client adapter for Metamist API."""

from metamist.graphql import gql, query, query_async


class GraphQLClient:
    """Adapter for GraphQL operations."""

    # Query definitions
    QUERY_DATASET_SGS = gql(
        """
        query DatasetData($dataset: String!, $seqTypes: [String!], $seqTechs: [String!], $seqPlatforms: [String!]) {
            project(name: $dataset) {
                sequencingGroups(type: {in_: $seqTypes}, technology: {in_: $seqTechs}, platform: {in_: $seqPlatforms}) {
                    id
                    type
                    technology
                    platform
                    sample {
                        id
                        externalIds
                        participant {
                            id
                            externalIds
                        }
                    }
                    assays {
                        id
                        meta
                    }
                }
            }
        }
        """
    )

    QUERY_SG_ANALYSES = gql(
        """
        query sgAnalyses($dataset: String!, $sgIds: [String!], $analysisTypes: [String!]) {
            project(name: $dataset) {
                sequencingGroups(id: {in_: $sgIds}) {
                    id
                    analyses(status: {eq: COMPLETED}, type: {in_: $analysisTypes}) {
                        id
                        meta
                        output
                        outputs
                        type
                        timestampCompleted
                    }
                }
            }
        }
        """
    )

    QUERY_AUDIT_DELETION_ANALYSES = gql(
        """
        query auditDeletions($dataset: String!) {
            project(name: $dataset) {
                analyses(status: {eq: COMPLETED}, type: {eq: "audit_deletion"}) {
                    id
                    timestampCompleted
                    output
                    meta
                }
            }
        }
        """
    )

    QUERY_COHORTS = gql(
        """
        query cohorts($dataset: String!) {
            project(name: $dataset) {
                cohorts {
                    id
                    name
                }
            }
        }
        """
    )

    QUERY_ENUMS = gql(
        """
        query enumsQuery {
            enum {
                analysisType
                sampleType
                sequencingType
                sequencingPlatform
                sequencingTechnology
            }
        }
        """
    )

    CREATE_COHORT_MUTATION = gql(
        """
        mutation createCohortMutation($dataset: String!, $name: String!) {
          cohort {
            createCohortFromCriteria(
              cohortSpec: {name: $name, description: "all_sgs_audit"}
              project: $dataset
              cohortCriteria: {projects: [$dataset]}
            ) {
              id
            }
          }
        }
        """
    )

    CREATE_ANALYSIS_MUTATION = gql(
        """
        mutation createAnalysisMutation($dataset: String!, $type: String!, $cohortIds: [String!], $output: String!, $meta: JSON!) {
          analysis {
            createAnalysis(analysis: {type: $type, status: COMPLETED, cohortIds: $cohortIds, output: $output, meta: $meta}, project: $dataset) {
                id
            }
          }
        }
        """
    )

    UPDATE_ANALYSIS_MUTATION = gql(
        """
        mutation updateAnalysisMutation($analysisId: Int!, $meta: JSON!) {
          analysis {
            updateAnalysis(analysis: {status: COMPLETED, meta: $meta}, analysisId: $analysisId) {
                id
            }
          }
        }
        """
    )

    def __init__(self):
        """Initialize GraphQL client."""
        self._enums_cache: dict[str, list[str]] | None = None

    async def get_enums(self) -> dict[str, list[str]]:
        """
        Get all available enum values from Metamist.

        Returns:
            Dictionary of enum types and their values
        """
        if self._enums_cache is None:
            result = await query_async(self.QUERY_ENUMS)
            self._enums_cache = result['enum']
        return self._enums_cache

    async def get_enum_values(self, enum_type: str) -> list[str]:
        """
        Get values for a specific enum type.

        Args:
            enum_type: The enum type to get values for

        Returns:
            List of enum values
        """
        enums = await self.get_enums()

        # Map common names to GraphQL field names
        enum_map = {
            'analysis_type': 'analysisType',
            'sample_type': 'sampleType',
            'sequencing_type': 'sequencingType',
            'sequencing_platform': 'sequencingPlatform',
            'sequencing_technology': 'sequencingTechnology',
        }

        graphql_field = enum_map.get(enum_type, enum_type)

        if graphql_field not in enums:
            raise KeyError(f'Enum {enum_type} not found in Metamist GraphQL API')

        return enums[graphql_field]

    async def get_sequencing_groups(
        self,
        dataset: str,
        sequencing_types: list[str],
        sequencing_technologies: list[str],
        sequencing_platforms: list[str],
    ) -> list[dict]:
        """
        Fetch sequencing groups from Metamist.

        Args:
            dataset: Dataset name
            sequencing_types: List of sequencing types to filter
            sequencing_technologies: List of sequencing technologies to filter
            sequencing_platforms: List of sequencing platforms to filter

        Returns:
            List of sequencing group dictionaries
        """
        result = await query_async(
            self.QUERY_DATASET_SGS,
            {
                'dataset': dataset,
                'seqTypes': sequencing_types,
                'seqTechs': sequencing_technologies,
                'seqPlatforms': sequencing_platforms,
            },
        )
        return result['project']['sequencingGroups']

    async def get_analyses_for_sequencing_groups(
        self, dataset: str, sg_ids: list[str], analysis_types: list[str]
    ) -> list[dict]:
        """
        Fetch analyses for given sequencing groups.

        Args:
            dataset: Dataset name
            sg_ids: List of sequencing group IDs
            analysis_types: List of analysis types to filter

        Returns:
            List of sequencing groups with their analyses
        """
        result = await query_async(
            self.QUERY_SG_ANALYSES,
            {
                'dataset': dataset,
                'sgIds': sg_ids,
                'analysisTypes': analysis_types,
            },
        )
        return result['project']['sequencingGroups']

    def get_audit_deletion_analyses(self, dataset: str) -> list[dict]:
        """
        Fetch completed audit deletion analyses for a dataset.

        Args:
            dataset: Dataset name

        Returns:
            List of audit deletion analyses
        """
        result = query(
            self.QUERY_AUDIT_DELETION_ANALYSES,
            {
                'dataset': dataset,
            },
        )
        return result['project']['analyses']

    def get_dataset_cohort(self, dataset: str, name: str) -> str:
        """
        Fetch or create the cohort associated with a dataset.

        Args:
            dataset: Dataset name

        Returns:
            Cohort ID
        """
        results = query(
            self.QUERY_COHORTS,
            {
                'dataset': dataset,
            },
        )
        for result in results['project']['cohorts']:
            if result['name'] == name:
                return result['id']

        # If not found, create the cohort
        result = query(
            self.CREATE_COHORT_MUTATION,
            {
                'dataset': dataset,
                'name': name,
            },
        )
        return result['cohort']['createCohortFromCriteria']['id']

    def create_audit_deletion_analysis(
        self, dataset: str, cohort_ids: list[str], output: str, meta: dict
    ) -> int:
        """
        Create a new audit deletion analysis.

        Args:
            dataset: Dataset name
            analysis_type: Type of analysis
            cohort_ids: String of cohort ID
            output: Output location
            meta: Metadata for the analysis

        Returns:
            Analysis ID
        """
        result = query(
            self.CREATE_ANALYSIS_MUTATION,
            {
                'dataset': dataset,
                'type': 'audit_deletion',
                'cohortIds': cohort_ids,
                'output': output,
                'meta': meta,
            },
        )
        return result['analysis']['createAnalysis']['id']

    def update_audit_deletion_analysis(self, analysis_id: int, meta: dict) -> int:
        """
        Update an existing audit deletion analysis.

        Args:
            analysis_id: Analysis ID
            meta: Updated metadata for the analysis

        Returns:
            Updated Analysis ID
        """
        result = query(
            self.UPDATE_ANALYSIS_MUTATION,
            {
                'analysisId': analysis_id,
                'meta': meta,
            },
        )
        return result['analysis']['updateAnalysis']['id']



"""Business logic services."""

from .file_matcher import (
    FileMatcher,
    ChecksumMatcher,
    FilenameSizeMatcher,
    CompositeFileMatcher,
    FileMatchingService,
)
from .audit_analyzer import AuditAnalyzer
from .reporter import Reporter
from .audit_logging import BucketAuditLogger

__all__ = [
    # File matching
    'FileMatcher',
    'ChecksumMatcher',
    'FilenameSizeMatcher',
    'CompositeFileMatcher',
    'FileMatchingService',
    # Core services
    'AuditAnalyzer',
    'Reporter',
    'BucketAuditLogger',
]



# pylint: disable=invalid-overridden-method
import os
from test.testbase import DbIsolatedTest, run_as_sync

from testcontainers.core.container import DockerContainer

from db.python.layers.analysis import AnalysisLayer
from db.python.layers.assay import AssayLayer
from db.python.layers.sample import SampleLayer
from db.python.layers.sequencing_group import SequencingGroupLayer
from db.python.tables.project import ProjectPermissionsTable
from models.enums import AnalysisStatus
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AssayUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from models.models.analysis import AnalysisInternal


class TestOutputFiles(DbIsolatedTest):
    """Test sample class"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()

        absolute_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
        gcs = (
            DockerContainer('fsouza/fake-gcs-server')
            .with_bind_ports(4443, 4443)
            .with_volume_mapping(
                absolute_path,
                '/data',
            )
            .with_command('-scheme http')
        )
        gcs.start()
        self.gcs = gcs

        self.sl = SampleLayer(self.connection)
        self.sgl = SequencingGroupLayer(self.connection)
        self.asl = AssayLayer(self.connection)
        self.al = AnalysisLayer(self.connection)

        sample = await self.sl.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                meta={'meta': 'meta ;)'},
                active=True,
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        sample_id=None,
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'sequencing_type': 'genome',
                                    'sequencing_technology': 'short-read',
                                    'sequencing_platform': 'illumina',
                                },
                            )
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        sample_id=None,
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'sequencing_type': 'exome',
                                    'sequencing_technology': 'short-read',
                                    'sequencing_platform': 'illumina',
                                },
                            )
                        ],
                    ),
                ],
            )
        )
        self.sample_id = sample.id
        assert sample.sequencing_groups
        self.genome_sequencing_group_id = sample.sequencing_groups[0].id
        self.exome_sequencing_group_id = sample.sequencing_groups[self.project_id].id

    def tearDown(self) -> None:
        if self.gcs:
            self.gcs.stop()
        return super().tearDown()

    def check_outputs_fields(self, outputs: dict, output_file_data: dict):
        """Check the fields of the output file"""
        if outputs:
            self.assertEqual(outputs['path'], output_file_data['path'])
            self.assertEqual(outputs['basename'], output_file_data['basename'])
            self.assertEqual(outputs['dirname'], output_file_data['dirname'])
            self.assertEqual(outputs['nameroot'], output_file_data['nameroot'])
            self.assertEqual(outputs['nameext'], output_file_data['nameext'])
            self.assertEqual(
                outputs['file_checksum'], output_file_data['file_checksum']
            )
            self.assertEqual(outputs['size'], output_file_data['size'])
            self.assertEqual(outputs['valid'], output_file_data['valid'])
            self.assertEqual(
                len(outputs['secondary_files']),
                len(output_file_data['secondary_files']),
            )

    @run_as_sync
    async def test_output_str(self):
        """Test how the output(s) behave when you create an analysis by passing in
        just the `output` field
        """

        output_path = 'FAKE://this_file_doesnt_exist.txt'

        # Create the analysis first
        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
                output=output_path,
            )
        )

        # Query the analysis object
        analysis = await self.al.get_analysis_by_id(analysis_id)
        assert analysis
        self.assertEqual(analysis.output, output_path)
        self.assertEqual(analysis.outputs, output_path)

    @run_as_sync
    async def test_gs_output_path(self):
        """Test how the output(s) behave when you create an analysis by passing in
        just the `output` field
        """
        output_path = 'gs://fakegcs/file1.txt'
        output_file_data = {
            'path': 'gs://fakegcs/file1.txt',
            'basename': 'file1.txt',
            'dirname': 'gs://fakegcs/',
            'nameroot': 'file1',
            'nameext': '.txt',
            'file_checksum': 'DG+fhg==',
            'size': 19,
            'valid': True,
            'secondary_files': {},
        }
        # Create the analysis first
        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
                output=output_path,
            )
        )

        # Query the analysis object
        analysis = await self.al.get_analysis_by_id(analysis_id)
        assert analysis

        assert isinstance(analysis.output, str)
        assert isinstance(analysis.outputs, dict)

        self.assertEqual(analysis.output, output_path)
        self.check_outputs_fields(analysis.outputs, output_file_data)

    @run_as_sync
    async def test_create_with_str_on_outputs(self):
        """This should test creating an Analysis by passing a string to the outputs field.
        The test should fail as we don't want to be passing string to this field.
        """

        output_path = 'gs://fakegcs/file1.txt'
        output_file_data = {
            'path': 'gs://fakegcs/file1.txt',
            'basename': 'file1.txt',
            'dirname': 'gs://fakegcs/',
            'nameroot': 'file1',
            'nameext': '.txt',
            'file_checksum': 'DG+fhg==',
            'size': 19,
            'valid': True,
            'secondary_files': {},
        }

        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
                outputs=output_path,
            )
        )

        analysis = await self.al.get_analysis_by_id(analysis_id)
        assert analysis

        assert isinstance(analysis.output, str)
        assert isinstance(analysis.outputs, dict)

        self.assertEqual(analysis.output, output_path)
        self.check_outputs_fields(analysis.outputs, output_file_data)

    @run_as_sync
    async def test_dict_with_outputs(self):
        """Should test creating Analysis object with a dictionary as the outputs field"""

        outputs = {
            'cram': {
                'basename': 'gs://fakegcs/file2.cram',
                'secondary_files': {
                    'meta': {'basename': 'gs://fakegcs/file2.cram.meta'},
                    'ext': {'basename': 'gs://fakegcs/file2.cram.ext'},
                },
            },
        }

        output_file_data = {
            'cram': {
                'path': 'gs://fakegcs/file2.cram',
                'basename': 'file2.cram',
                'dirname': 'gs://fakegcs/',
                'nameroot': 'file2',
                'nameext': '.cram',
                'file_checksum': 'sl7SXw==',
                'size': 20,
                'meta': None,
                'valid': True,
                'secondary_files': {
                    'meta': {
                        'path': 'gs://fakegcs/file2.cram.meta',
                        'basename': 'file2.cram.meta',
                        'dirname': 'gs://fakegcs/',
                        'nameroot': 'file2.cram',
                        'nameext': '.meta',
                        'file_checksum': 'af/YSw==',
                        'size': 17,
                        'meta': None,
                        'valid': True,
                        'secondary_files': {},
                    },
                    'ext': {
                        'path': 'gs://fakegcs/file2.cram.ext',
                        'basename': 'file2.cram.ext',
                        'dirname': 'gs://fakegcs/',
                        'nameroot': 'file2.cram',
                        'nameext': '.ext',
                        'file_checksum': 'gb1EbA==',
                        'size': 21,
                        'meta': None,
                        'valid': True,
                        'secondary_files': {},
                    },
                },
            }
        }

        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
                outputs=outputs,
            )
        )

        analysis = await self.al.get_analysis_by_id(analysis_id)
        assert analysis

        assert isinstance(analysis.output, str)
        assert isinstance(analysis.outputs, dict)

        self.assertEqual(analysis.output, '')
        self.assertIn('cram', analysis.outputs)
        self.assertIn('secondary_files', analysis.outputs['cram'])
        self.assertIn('meta', analysis.outputs['cram']['secondary_files'])
        self.assertIn('ext', analysis.outputs['cram']['secondary_files'])

        # Check for each field against the output file data and also check each secondary file
        self.check_outputs_fields(analysis.outputs['cram'], output_file_data['cram'])
        self.check_outputs_fields(
            analysis.outputs['cram']['secondary_files']['meta'],
            output_file_data['cram']['secondary_files']['meta'],  # type: ignore [index]
        )
        self.check_outputs_fields(
            analysis.outputs['cram']['secondary_files']['ext'],
            output_file_data['cram']['secondary_files']['ext'],  # type: ignore [index]
        )

    @run_as_sync
    async def test_project_deletion(self):
        """Test ProjectPermissionsTable.delete_project_data's effect on analysis outputs and files"""

        outputs = {
            'cram': {
                'basename': 'gs://fakegcs/file3.cram',
                'secondary_files': {
                    'meta': {'basename': 'gs://fakegcs/file3.cram.meta'},
                    'ext': {'basename': 'gs://fakegcs/file3.cram.ext'},
                },
            },
        }

        await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
                outputs=outputs,
            )
        )

        self.assertEqual(await self.row_count('analysis_outputs'), 3)
        self.assertEqual(await self.row_count('output_file'), 3)

        pttable = ProjectPermissionsTable(self.connection)
        project = self.project_id_map[self.project_id]
        self.assertTrue(await pttable.delete_project_data(project))

        self.assertEqual(await self.row_count('analysis_outputs'), 0)
        self.assertEqual(await self.row_count('output_file'), 0)



from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.sample import SampleLayer
from models.models import PRIMARY_EXTERNAL_ORG, SampleUpsertInternal


class TestChangelog(DbIsolatedTest):
    """Test audit_log"""

    @run_as_sync
    async def test_insert_sample(self):
        """
        Test inserting a sample, and check that the audit_log_id reflects the current
        change
        """
        slayer = SampleLayer(self.connection)
        sample = await slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        sample_cl_id = await self.connection.connection.fetch_val(
            'SELECT audit_log_id FROM sample WHERE id = :sid', {'sid': sample.id}
        )

        self.assertEqual(await self.audit_log_id(), sample_cl_id)



# pylint: disable=ungrouped-imports, too-many-lines

from db.python.filters.generic import GenericFilter
from db.python.layers.cohort import CohortLayer
from db.python.layers.family import FamilyLayer
from db.python.tables.cohort import CohortFilter, CohortTemplateFilter
from db.python.tables.project import ProjectPermissionsTable
from models.models.project import ProjectMemberRole
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AnalysisInternal,
    AssayUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from models.models.cohort import CohortCriteriaInternal, CohortTemplateInternal
from models.models.participant import ParticipantUpsertInternal
from models.utils.sample_id_format import sample_id_transform_to_raw
from models.utils.sequencing_group_id_format import sequencing_group_id_transform_to_raw
from test.testbase import DbIsolatedTest, run_as_sync
from db.python.layers.analysis import AnalysisLayer
from db.python.layers.assay import AssayLayer
from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from db.python.layers.sequencing_group import SequencingGroupLayer
from models.enums import AnalysisStatus
from api.graphql.mutations.analysis import AnalysisStatusType

GROUP_NAME_PROJECT_CREATORS = 'project-creators'

# region ANALYSIS MUTATIONS

CREATE_ANALYSIS_MUTATION = """
    mutation createAnalysis($project: String!, $sequencingGroupIds: [String!], $status: AnalysisStatus!, $type: String!) {
        analysis {
            createAnalysis(project: $project, analysis: {
                type: $type,
                status: $status,
                meta:{},
                sequencingGroupIds: $sequencingGroupIds,
            }){
                id
                status
                meta
                sequencingGroups {
                    id
                }
                type
            }
        }
    }
"""

UPDATE_ANALYSIS_MUTATION = """
    mutation updateAnalysis($analysisId: Int!, $status: AnalysisStatus!, $meta: JSON) {
        analysis {
            updateAnalysis(analysisId: $analysisId, analysis: {
                status: $status,
                meta: $meta,
            }) {
                id
                status
                meta
                sequencingGroups {
                    id
                }
                type
            }
        }
    }
"""

# endregion ANALYSIS MUTATIONS

# region ASSAY MUTATIONS

CREATE_ASSAY_MUTATION = """
    mutation createAssay($type: String!, $meta: JSON, $externalIds: JSON, $sampleId: String) {
        assay {
            createAssay(assay: {
                type: $type,
                meta: $meta,
                externalIds: $externalIds,
                sampleId: $sampleId,
            }){
                id
                type
                meta
                externalIds
                sample {
                    id
                }
            }
        }
    }
"""

UPDATE_ASSAY_MUTATION = """
    mutation updateAssay($assayId: Int!, $type: String!, $meta: JSON, $externalIds: JSON, $sampleId: String) {
        assay {
            updateAssay(assay: {
                id: $assayId,
                type: $type,
                meta: $meta,
                externalIds: $externalIds,
                sampleId: $sampleId,
            }) {
                id
                type
                meta
                externalIds
                sample {
                    id
                }
            }
        }
    }
"""

# endregion ASSAY MUTATIONS

# region COHORT MUTATIONS

CREATE_COHORT_FROM_CRITERIA_MUTATION = """
    mutation CreateCohortFromCriteria($project: String!, $cohortSpec: CohortBodyInput!, $cohortCriteria: CohortCriteriaInput!, $dryRun: Boolean) {
        cohort{
            createCohortFromCriteria(
                project: $project
                cohortSpec: $cohortSpec
                cohortCriteria: $cohortCriteria
                dryRun: $dryRun
            ) {
                id
                name
                description
                author
            }
        }
    }
"""

CREATE_COHORT_TEMPLATE_MUTATION = """
    mutation createCohortTemplate($project: String!, $template: CohortTemplateInput!) {
        cohort {
            createCohortTemplate(
                project: $project
                template: $template
            ) {
                id
                name
                description
                criteria
            }
        }
    }
"""
# endregion COHORT MUTATIONS

# region FAMILY MUTATIONS
UPDATE_FAMILY_MUTATION = """
    mutation updateFamily($family: FamilyUpdateInput!) {
        family {
            updateFamily(family: $family) {
                id
                externalIds
                description
                codedPhenotype
            }
        }
    }
"""
# endregion FAMILY MUTATIONS

# region PARTICIPANT MUTATIONS
UPDATE_PARTICIPANT_MUTATION = """
    mutation updateParticipant($participantId: Int!, $participant: ParticipantUpsertInput!) {
        participant {
            updateParticipant(participantId: $participantId, participant: $participant) {
                id
                externalIds
                reportedSex
                reportedGender
                karyotype
                samples {
                    id
                    type
                    meta
                    externalIds
                }
            }
        }
    }
"""

UPSERT_PARTICIPANTS_MUTATION = """
    mutation upsertParticipants($project: String!, $participants: [ParticipantUpsertInput!]!) {
        participant {
            upsertParticipants(project: $project, participants: $participants) {
                id
                externalIds
                reportedSex
                reportedGender
                karyotype
                samples {
                    id
                    type
                    meta
                    externalIds
                }
            }
        }
    }
"""

UPDATE_PARTICIPANT_FAMILY_MUTATION = """
    mutation updateParticipantFamily($participantId: Int!, $oldFamilyId: Int!, $newFamilyId: Int!) {
        participant {
            updateParticipantFamily(participantId: $participantId, oldFamilyId: $oldFamilyId, newFamilyId: $newFamilyId) {
                familyId
                participantId
            }
        }
    }
"""
# endregion PARTICIPANT MUTATIONS

# region PROJECT MUTATIONS
CREATE_PROJECT_MUTATION = """
    mutation createProject($name: String!, $dataset: String!, $createTestProject: Boolean!) {
        project {
            createProject(name: $name, dataset: $dataset, createTestProject: $createTestProject) {
                id
                name
                dataset
                meta
            }
        }
    }
"""

UPDATE_PROJECT_MUTATION = """
    mutation updateProject($project: String!, $projectUpdateModel: JSON!) {
        project {
            updateProject(project: $project, projectUpdateModel: $projectUpdateModel) {
                id
                name
                dataset
                meta
            }
        }
    }
"""

UPDATE_PROJECT_MEMBERS_MUTATION = """
    mutation updateProjectMembers($project: String!, $members: [ProjectMemberUpdateInput!]!) {
        project {
            updateProjectMembers(project: $project, members: $members) {
                id
                name
                dataset
                meta
                roles
            }
        }
    }
"""
# endregion PROJECT MUTATIONS

# region SAMPLE MUTATIONS
CREATE_SAMPLE_MUTATION = """
    mutation createSample($project: String!, $sample: SampleUpsertInput!) {
        sample {
            createSample(project: $project, sample: $sample) {
                id
                externalIds
                type
                active
                meta
            }
        }
    }
"""

UPSERT_SAMPLES_MUTATION = """
    mutation upsertSamples($project: String!, $samples: [SampleUpsertInput!]!) {
        sample {
            upsertSamples(project: $project, samples: $samples) {
                id
                externalIds
                type
                active
                meta
            }
        }
    }
"""

UPDATE_SAMPLE_MUTATION = """
    mutation updateSample($sample: SampleUpsertInput!) {
        sample {
            updateSample(sample: $sample) {
                id
                externalIds
                type
                active
                meta
            }
        }
    }
"""
# endregion SAMPLE MUTATIONS

# region SEQUENCING GROUP MUTATIONS
UPDATE_SEQUENCING_GROUP_MUTATION = """
    mutation updateSequencingGroup($project: String!, $sequencingGroup: SequencingGroupMetaUpdateInput!) {
        sequencingGroup {
            updateSequencingGroup(project: $project, sequencingGroup: $sequencingGroup) {
                id
                type
                meta
                externalIds
                sample {
                    id
                }
            }
        }
    }
"""
# endregion SEQUENCING GROUP MUTATIONS


class TestMutations(DbIsolatedTest):
    """Test sample class"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()  # type: ignore [call-arg]
        self.cl = CohortLayer(self.connection)
        self.sl = SampleLayer(self.connection)
        self.sgl = SequencingGroupLayer(self.connection)
        self.asl = AssayLayer(self.connection)
        self.al = AnalysisLayer(self.connection)
        self.pl = ParticipantLayer(self.connection)
        self.fl = FamilyLayer(self.connection)
        self.ppt = ProjectPermissionsTable(self.connection)

        self.family_id = await self.fl.create_family(external_ids={'forg': 'FAM01'})
        self.family_id_2 = await self.fl.create_family(external_ids={'forg': 'FAM02'})

        sample = await self.sl.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                meta={'meta': 'meta ;)'},
                active=True,
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        sample_id=None,
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'sequencing_type': 'genome',
                                    'sequencing_technology': 'short-read',
                                    'sequencing_platform': 'illumina',
                                },
                            )
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        sample_id=None,
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'sequencing_type': 'exome',
                                    'sequencing_technology': 'short-read',
                                    'sequencing_platform': 'illumina',
                                },
                            )
                        ],
                    ),
                ],
            )
        )
        self.sample_id = sample.id
        self.external_sample_id = sample.to_external().id
        self.genome_sequencing_group_id = sample.sequencing_groups[0].id  # type: ignore [arg-type]
        self.genome_sequencing_group_id_external = (
            sample.sequencing_groups[0].to_external().id  # type: ignore [arg-type]
        )
        self.exome_sequencing_group_id = sample.sequencing_groups[self.project_id].id  # type: ignore [arg-type]
        self.exome_sequencing_group_id_external = (
            sample.sequencing_groups[self.project_id].to_external().id  # type: ignore [arg-type]
        )
        self.participant = await self.pl.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX01'},
                reported_sex=2,
                samples=[sample],
            )
        )
        self.pat_id = (
            await self.pl.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01_pat'}, reported_sex=1
                )
            )
        ).id
        self.mat_id = (
            await self.pl.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01_mat'}, reported_sex=2
                )
            )
        ).id

        assert self.participant.id
        assert self.pat_id
        assert self.mat_id

        await self.pl.add_participant_to_family(
            family_id=self.family_id,
            participant_id=self.participant.id,
            paternal_id=self.pat_id,
            maternal_id=self.mat_id,
            affected=2,
        )

        await self.connection.connection.execute(
            f"""
            INSERT INTO group_member(group_id, member)
            SELECT id, '{self.author}'
            FROM `group` WHERE name IN('project-creators', 'members-admin')
            """
        )

    # region ANALYSIS TESTS

    @run_as_sync
    async def test_create_analysis(self):
        """Test creating an analysis using the mutation and the API"""
        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_ANALYSIS_MUTATION,
                variables={
                    'project': self.project_name,
                    'sequencingGroupIds': [self.genome_sequencing_group_id_external],
                    'status': AnalysisStatusType.UNKNOWN.name,
                    'type': 'analysis-runner',
                },
            )
        )['analysis']['createAnalysis']

        aid = await self.al.create_analysis(
            project=self.project_id,
            analysis=AnalysisInternal(
                type='analysis-runner',
                status=AnalysisStatus.UNKNOWN,
                meta={},
                sequencing_group_ids=[self.genome_sequencing_group_id],  # type: ignore [arg-type]
            ),
        )

        api_result = (await self.al.get_analysis_by_id(aid)).to_external()

        self.assertEqual(
            api_result.type,
            mutation_result['type'],
        )
        self.assertEqual(
            api_result.status.name,
            mutation_result['status'],
        )
        self.assertEqual(
            api_result.sequencing_group_ids,
            [s['id'] for s in mutation_result['sequencingGroups']],
        )
        self.assertEqual(
            api_result.meta,
            mutation_result['meta'],
        )

    @run_as_sync
    async def test_update_analysis(self):
        """Test updating an analysis using the mutation and the API"""

        analysis = await self.al.create_analysis(
            project=self.project_id,
            analysis=AnalysisInternal(
                type='analysis-runner',
                status=AnalysisStatus.UNKNOWN,
                meta={},
                sequencing_group_ids=[self.genome_sequencing_group_id],  # type: ignore [arg-type]
            ),
        )
        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_ANALYSIS_MUTATION,
                variables={
                    'analysisId': analysis,
                    'status': AnalysisStatusType.COMPLETED.name,
                    'meta': {'test': 'test'},
                },
            )
        )['analysis']['updateAnalysis']

        await self.al.update_analysis(
            analysis_id=analysis,
            status=AnalysisStatus.COMPLETED,
            meta={'test': 'test'},
        )

        api_result = (await self.al.get_analysis_by_id(analysis)).to_external()

        self.assertEqual(api_result.status.name, mutation_result['status'])
        self.assertEqual(api_result.meta, mutation_result['meta'])

    # endregion ANALYSIS TESTS

    # region ASSAY TESTS
    @run_as_sync
    async def test_create_assay(self):
        """Test creating an assay using the mutation and the API"""
        default_sequencing_meta = {
            'sequencing_type': 'genome',
            'sequencing_platform': 'short-read',
            'sequencing_technology': 'illumina',
        }
        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_ASSAY_MUTATION,
                variables={
                    'project': self.project_name,
                    'externalIds': {'test1': 'test1'},
                    'sampleId': self.external_sample_id,
                    'type': 'sequencing',
                    'meta': {'test': 'test', **default_sequencing_meta},
                },
            )
        )['assay']['createAssay']

        api_result = (
            await self.asl.upsert_assay(
                assay=AssayUpsertInternal(
                    type='sequencing',
                    meta={'test': 'test', **default_sequencing_meta},
                    external_ids={'test': 'test'},
                    sample_id=self.sample_id,
                ),
            )
        ).to_external()

        self.assertEqual(
            api_result.type,
            mutation_result['type'],
        )
        self.assertEqual(
            api_result.meta,
            mutation_result['meta'],
        )
        self.assertEqual(
            api_result.sample_id,
            mutation_result['sample']['id'],
        )

    @run_as_sync
    async def test_update_assay(self):
        """Test updating an assay using the mutation and the API"""

        default_sequencing_meta = {
            'sequencing_type': 'genome',
            'sequencing_platform': 'short-read',
            'sequencing_technology': 'illumina',
        }

        assay_id = (
            (
                await self.asl.upsert_assay(
                    assay=AssayUpsertInternal(
                        type='sequencing',
                        meta={'test': 'test', **default_sequencing_meta},
                        external_ids={'test': 'test'},
                        sample_id=self.sample_id,
                    ),
                )
            )
            .to_external()
            .id
        )

        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_ASSAY_MUTATION,
                variables={
                    'project': self.project_name,
                    'assayId': assay_id,
                    'type': 'sequencing',
                    'meta': {'test': 'test2', **default_sequencing_meta},
                },
            )
        )['assay']['updateAssay']

        await self.asl.upsert_assay(
            AssayUpsertInternal(
                id=assay_id,  # type: ignore [arg-type]
                type='sequencing',
                meta={'test': 'test2', **default_sequencing_meta},
            )
        )

        api_result = (await self.asl.get_assay_by_id(assay_id)).to_external()  # type: ignore [arg-type]

        self.assertEqual(api_result.type, mutation_result['type'])
        self.assertEqual(api_result.meta, mutation_result['meta'])
        self.assertEqual(api_result.sample_id, mutation_result['sample']['id'])
        self.assertEqual(api_result.external_ids, mutation_result['externalIds'])

    # endregion ASSAY TESTS

    # region COHORT TESTS
    @run_as_sync
    async def test_create_cohort_from_criteria(self):
        """Test creating a cohort from criteria using the mutation and the API"""

        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_COHORT_FROM_CRITERIA_MUTATION,
                variables={
                    'project': self.project_name,
                    'cohortSpec': {
                        'name': 'TestCohort1',
                        'description': 'TestCohortDescription',
                        # 'templateId': cohort_template_id_format(tid),
                    },
                    'cohortCriteria': {
                        'projects': [self.project_name],
                        'sgIdsInternal': [self.genome_sequencing_group_id_external],
                        'excludedSgsInternal': [
                            self.exome_sequencing_group_id_external
                        ],
                        'sgTechnology': ['short-read'],
                        'sgPlatform': ['illumina'],
                        'sgType': ['genome'],
                        'sampleType': ['blood'],
                    },
                },
            )
        )['cohort']['createCohortFromCriteria']
        cohort = await self.cl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='TestCohortDescription',
            cohort_name='TestCohort2',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_ids_internal_raw=[self.genome_sequencing_group_id],  # type: ignore [arg-type]
                excluded_sgs_internal_raw=[self.exome_sequencing_group_id],  # type: ignore [arg-type]
                sg_technology=['short-read'],
                sg_platform=['illumina'],
                sg_type=['genome'],
                sample_type=['blood'],
            ),
        )
        api_result = (
            await self.cl.query(CohortFilter(id=GenericFilter(eq=cohort.cohort_id)))
        )[0]
        self.assertEqual(api_result.description, mutation_result['description'])
        self.assertEqual(api_result.author, mutation_result['author'])

    @run_as_sync
    async def test_create_cohort_template(self):
        """Test creating a cohort template"""

        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_COHORT_TEMPLATE_MUTATION,
                variables={
                    'project': self.project_name,
                    'template': {
                        'name': 'TestTemplate',
                        'description': 'TestCohortTemplateDescription',
                        'criteria': {
                            'projects': [self.project_name],
                            'sgIdsInternal': [self.genome_sequencing_group_id_external],
                            'excludedSgsInternal': [
                                self.exome_sequencing_group_id_external
                            ],
                            'sgTechnology': ['short-read'],
                            'sgPlatform': ['illumina'],
                            'sgType': ['genome'],
                            'sampleType': ['blood'],
                        },
                    },
                },
            )
        )['cohort']['createCohortTemplate']
        template_id = await self.cl.create_cohort_template(
            project=self.project_id,
            cohort_template=CohortTemplateInternal(
                id=None,
                name='TestTemplate',
                description='TestCohortTemplateDescription',
                project=self.project_id,
                criteria=CohortCriteriaInternal(
                    projects=[self.project_id],
                    sg_ids_internal_raw=[self.genome_sequencing_group_id],  # type: ignore [arg-type]
                    excluded_sgs_internal_raw=[self.exome_sequencing_group_id],  # type: ignore [arg-type]
                    sg_technology=['short-read'],
                    sg_platform=['illumina'],
                    sg_type=['genome'],
                    sample_type=['blood'],
                ),
            ),
        )
        api_result = (
            await self.cl.query_cohort_templates(
                CohortTemplateFilter(id=GenericFilter(eq=template_id))
            )
        )[0]
        self.assertEqual(api_result.description, mutation_result['description'])
        self.assertEqual(
            api_result.criteria.sample_type, mutation_result['criteria']['sample_type']
        )
        self.assertEqual(
            api_result.criteria.sg_platform, mutation_result['criteria']['sg_platform']
        )
        self.assertEqual(
            api_result.criteria.sg_technology,
            mutation_result['criteria']['sg_technology'],
        )
        self.assertEqual(
            api_result.criteria.sg_type, mutation_result['criteria']['sg_type']
        )
        self.assertEqual(api_result.name, mutation_result['name'])

    # endregion COHORT TESTS

    # region FAMILY TESTS
    @run_as_sync
    async def test_update_family(self):
        """Test updating a family using the mutation and the API"""
        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_FAMILY_MUTATION,
                variables={
                    'project': self.project_name,
                    'family': {
                        'id': self.family_id,
                        'externalIds': {PRIMARY_EXTERNAL_ORG: 'test'},
                        'description': 'test_family',
                        'codedPhenotype': 'test_family_phenotype',
                    },
                },
            )
        )['family']['updateFamily']

        await self.fl.update_family(
            id_=self.family_id,
            external_ids={'test': 'test'},
            description='test_family',
            coded_phenotype='test_family_phenotype',
        )

        api_result = (
            await self.fl.get_family_by_internal_id(self.family_id)
        ).to_external()  # type: ignore [arg-type]

        self.assertEqual(api_result.external_ids, mutation_result['externalIds'])
        self.assertEqual(api_result.description, mutation_result['description'])
        self.assertEqual(api_result.coded_phenotype, mutation_result['codedPhenotype'])

    # endregion FAMILY TESTS

    # region PARTICIPANT TESTS
    @run_as_sync
    async def test_update_participant(self):
        """Test updating a participant using the mutation and the API"""
        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_PARTICIPANT_MUTATION,
                variables={
                    'project': self.project_name,
                    'participantId': self.participant.id,
                    'participant': {
                        'id': self.participant.id,
                        'externalIds': {PRIMARY_EXTERNAL_ORG: 'test'},
                        'reportedSex': 2,
                        'reportedGender': 'female',
                        'karyotype': 'test_karyotype',
                        'samples': [
                            {
                                'id': self.external_sample_id,
                                'type': 'blood',
                                'meta': {'test': 'test'},
                                'externalIds': {'test': 'test'},
                            }
                        ],
                    },
                },
            )
        )['participant']['updateParticipant']

        api_result = (
            await self.pl.upsert_participant(
                ParticipantUpsertInternal(
                    id=self.participant.id,
                    external_ids={PRIMARY_EXTERNAL_ORG: 'test'},
                    reported_sex=2,
                    reported_gender='female',
                    karyotype='test_karyotype',
                    samples=[
                        SampleUpsertInternal(
                            id=self.sample_id,
                            type='blood',
                            meta={'test': 'test'},
                            external_ids={'test': 'test'},
                        )
                    ],
                )
            )
        ).to_external()

        self.assertEqual(api_result.external_ids, mutation_result['externalIds'])
        self.assertEqual(api_result.reported_sex, mutation_result['reportedSex'])
        self.assertEqual(api_result.reported_gender, mutation_result['reportedGender'])
        self.assertEqual(api_result.karyotype, mutation_result['karyotype'])
        self.assertEqual(api_result.samples[0].id, mutation_result['samples'][0]['id'])  # type: ignore [arg-type]

    @run_as_sync
    async def test_upsert_participants(self):
        """Test upserting a list of participants using the mutation and the API. This inserts a new participant and updates an existing one."""
        mutation_result = (
            await self.run_graphql_query_async(
                UPSERT_PARTICIPANTS_MUTATION,
                variables={
                    'project': self.project_name,
                    'participants': [
                        {
                            'id': self.participant.id,
                            'externalIds': {PRIMARY_EXTERNAL_ORG: 'EX01'},
                            'reportedSex': 2,
                            'reportedGender': 'female',
                            'karyotype': 'test_karyotype',
                            'samples': [
                                {
                                    'id': self.external_sample_id,
                                    'type': 'blood',
                                    'meta': {'test': 'test'},
                                    'externalIds': {'test': 'test'},
                                }
                            ],
                        },
                        {
                            'externalIds': {PRIMARY_EXTERNAL_ORG: 'EX02_pat'},
                            'reportedSex': 1,
                            'reportedGender': 'female',
                            'karyotype': 'test_karyotype',
                            'samples': [
                                {
                                    'id': self.external_sample_id,
                                    'type': 'blood',
                                    'meta': {'test': 'test'},
                                    'externalIds': {'test': 'test'},
                                }
                            ],
                        },
                    ],
                },
            )
        )['participant']['upsertParticipants']

        api_result = await self.pl.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01'},
                    reported_sex=2,
                    samples=[
                        SampleUpsertInternal(
                            id=self.sample_id,
                            type='blood',
                            meta={'test': 'test'},
                            external_ids={'test': 'test'},
                        )
                    ],
                    id=self.participant.id,
                    reported_gender='female',
                    karyotype='test_karyotype',
                ),
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX03_pat'},
                    reported_sex=1,
                    samples=[
                        SampleUpsertInternal(
                            id=self.sample_id,
                            type='blood',
                            meta={'test': 'test'},
                            external_ids={'test': 'test'},
                        )
                    ],
                    reported_gender='female',
                    karyotype='test_karyotype',
                    id=None,
                ),
            ]
        )

        api_result = [p.to_external() for p in api_result]
        self.assertEqual(api_result[0].id, mutation_result[0]['id'])
        self.assertEqual(api_result[0].external_ids, mutation_result[0]['externalIds'])
        self.assertEqual(api_result[0].reported_sex, mutation_result[0]['reportedSex'])
        self.assertEqual(
            api_result[0].reported_gender, mutation_result[0]['reportedGender']
        )
        self.assertEqual(api_result[0].karyotype, mutation_result[0]['karyotype'])

        self.assertEqual(api_result[1].reported_sex, mutation_result[1]['reportedSex'])
        self.assertEqual(
            api_result[1].reported_gender, mutation_result[1]['reportedGender']
        )
        self.assertEqual(api_result[1].karyotype, mutation_result[1]['karyotype'])

    @run_as_sync
    async def test_update_participant_family(self):
        """Test updating a participants family data"""
        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_PARTICIPANT_FAMILY_MUTATION,
                variables={
                    'project': self.project_name,
                    'participantId': self.participant.id,
                    'oldFamilyId': self.family_id,
                    'newFamilyId': self.family_id_2,
                },
            )
        )['participant']['updateParticipantFamily']

        api_result = (
            await self.fl.get_family_participants_by_family_ids([self.family_id_2])
        )[self.family_id_2][0]

        self.assertEqual(api_result.individual_id, mutation_result['participantId'])
        self.assertEqual(api_result.family_id, mutation_result['familyId'])

    # endregion PARTICIPANT TESTS

    # region PROJECT TESTS
    @run_as_sync
    async def test_create_project(self):
        """Test creating a project using the mutation and the API"""
        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_PROJECT_MUTATION,
                variables={
                    'name': 'test_project',
                    'dataset': 'test_dataset',
                    'createTestProject': True,
                },
            )
        )['project']['createProject']

        api_result = list(
            self.connection.get_and_check_access_to_projects_for_names(
                [mutation_result['name'], mutation_result['name'] + '-test'],
                allowed_roles={
                    ProjectMemberRole.project_admin,
                    ProjectMemberRole.writer,
                },
            )
        )

        self.assertEqual(api_result[0].name, mutation_result['name'])
        self.assertEqual(api_result[0].dataset, mutation_result['dataset'])
        self.assertEqual(api_result[0].meta, mutation_result['meta'])

        self.assertEqual(api_result[1].name, mutation_result['name'] + '-test')
        self.assertEqual(api_result[1].dataset, mutation_result['dataset'])
        self.assertEqual(api_result[1].meta, mutation_result['meta'])

    @run_as_sync
    async def test_update_project(self):
        """Test updating a project using the mutation and the API"""
        create_project_result = (
            await self.run_graphql_query_async(
                CREATE_PROJECT_MUTATION,
                variables={
                    'name': 'new_test_project',
                    'dataset': 'test_dataset',
                    'createTestProject': False,
                },
            )
        )['project']['createProject']

        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_PROJECT_MUTATION,
                variables={
                    'project': create_project_result['name'],
                    'projectUpdateModel': {
                        'meta': {'test': 'test'},
                    },
                },
            )
        )['project']['updateProject']

        await self.connection.refresh_projects()

        api_result = list(
            self.connection.get_and_check_access_to_projects_for_names(
                [mutation_result['name']],
                allowed_roles={
                    ProjectMemberRole.project_admin,
                    ProjectMemberRole.writer,
                },
            )
        )[0]

        self.assertEqual(api_result.name, mutation_result['name'])
        self.assertEqual(api_result.dataset, mutation_result['dataset'])
        self.assertEqual(api_result.meta, {'test': 'test'})

    @run_as_sync
    async def test_update_project_members(self):
        """Test updating project members using the mutation and the API"""
        create_project_result = (
            await self.run_graphql_query_async(
                CREATE_PROJECT_MUTATION,
                variables={
                    'name': 'new_test_project2',
                    'dataset': 'test_dataset',
                    'createTestProject': False,
                },
            )
        )['project']['createProject']

        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_PROJECT_MEMBERS_MUTATION,
                variables={
                    'project': create_project_result['name'],
                    'members': [
                        {
                            'member': 'testuser',
                            'roles': ['reader', 'writer'],
                        }
                    ],
                },
            )
        )['project']['updateProjectMembers']

        api_result = list(
            self.connection.get_and_check_access_to_projects_for_names(
                [mutation_result['name']],
                allowed_roles={
                    ProjectMemberRole.project_member_admin,
                },
            )
        )[0]

        self.assertEqual(api_result.name, mutation_result['name'])
        self.assertEqual(api_result.dataset, mutation_result['dataset'])
        self.assertEqual(api_result.meta, mutation_result['meta'])
        self.assertEqual(
            [role.value for role in api_result.roles], mutation_result['roles']
        )

    # endregion PROJECT TESTS

    # region SAMPLE TESTS
    @run_as_sync
    async def test_create_sample(self):
        """Test creating a sample using the mutation and the API"""
        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_SAMPLE_MUTATION,
                variables={
                    'project': self.project_name,
                    'sample': {
                        'type': 'blood',
                        'meta': {'test': 'test'},
                        'externalIds': {PRIMARY_EXTERNAL_ORG: 'Test10'},
                        'active': True,
                    },
                },
            )
        )['sample']['createSample']

        api_result = await self.sl.get_sample_by_id(
            sample_id_transform_to_raw(mutation_result['id'])
        )

        self.assertEqual(
            api_result.type,
            'blood',
        )
        self.assertEqual(
            api_result.meta,
            {'test': 'test'},
        )
        self.assertEqual(
            api_result.external_ids,
            {PRIMARY_EXTERNAL_ORG: 'Test10'},
        )
        self.assertEqual(
            api_result.active,
            True,
        )

    @run_as_sync
    async def test_upsert_samples(self):
        """Test upserting a list of samples using the mutation and the API. This inserts a new sample and updates an existing one."""
        mutation_result = (
            await self.run_graphql_query_async(
                UPSERT_SAMPLES_MUTATION,
                variables={
                    'project': self.project_name,
                    'samples': [
                        {
                            'id': self.external_sample_id,
                            'type': 'blood',
                            'meta': {'test': 'test'},
                            'externalIds': {PRIMARY_EXTERNAL_ORG: 'Test10'},
                            'active': True,
                        },
                        {
                            'externalIds': {PRIMARY_EXTERNAL_ORG: 'Test11'},
                            'type': 'saliva',
                            'meta': {'test': 'test'},
                            'active': True,
                        },
                    ],
                },
            )
        )['sample']['upsertSamples']

        api_result = await self.sl.get_samples_by(
            sample_ids=[sample_id_transform_to_raw(s['id']) for s in mutation_result]
        )

        self.assertEqual(len(api_result), 2)
        self.assertEqual(
            api_result[0].type,
            'blood',
        )
        self.assertEqual(
            api_result[0].external_ids,
            {PRIMARY_EXTERNAL_ORG: 'Test10'},
        )
        self.assertEqual(
            api_result[0].active,
            True,
        )
        self.assertEqual(
            api_result[1].type,
            'saliva',
        )
        self.assertEqual(
            api_result[1].meta,
            {'test': 'test'},
        )
        self.assertEqual(
            api_result[1].external_ids,
            {PRIMARY_EXTERNAL_ORG: 'Test11'},
        )
        self.assertEqual(
            api_result[1].active,
            True,
        )

    @run_as_sync
    async def test_update_sample(self):
        """Test updating a sample using the mutation and the API"""
        create_sample_result = (
            await self.run_graphql_query_async(
                CREATE_SAMPLE_MUTATION,
                variables={
                    'project': self.project_name,
                    'sample': {
                        'type': 'blood',
                        'meta': {'test': 'test'},
                        'externalIds': {PRIMARY_EXTERNAL_ORG: 'Test10'},
                        'active': True,
                    },
                },
            )
        )['sample']['createSample']

        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_SAMPLE_MUTATION,
                variables={
                    'sample': {
                        'id': create_sample_result['id'],
                        'type': 'saliva',
                        'meta': {'test': 'test'},
                        'externalIds': {PRIMARY_EXTERNAL_ORG: 'Test11'},
                        'active': True,
                    },
                },
            )
        )['sample']['updateSample']

        api_result = await self.sl.get_sample_by_id(
            sample_id_transform_to_raw(mutation_result['id'])
        )

        self.assertEqual(
            api_result.type,
            'saliva',
        )
        self.assertEqual(
            api_result.meta,
            {'test': 'test'},
        )
        self.assertEqual(
            api_result.external_ids,
            {PRIMARY_EXTERNAL_ORG: 'Test11'},
        )
        self.assertEqual(
            api_result.active,
            True,
        )

    # endregion SAMPLE TESTS

    # region SEQUENCING GROUP TESTS
    @run_as_sync
    async def test_update_sequencing_group(self):
        """Test updating a sequencing group using the mutation and the API"""
        mutation_result = (
            await self.run_graphql_query_async(
                UPDATE_SEQUENCING_GROUP_MUTATION,
                variables={
                    'project': self.project_name,
                    'sequencingGroup': {
                        'id': self.genome_sequencing_group_id_external,
                        'meta': {'test': 'test'},
                    },
                },
            )
        )['sequencingGroup']['updateSequencingGroup']

        api_result = await self.sgl.get_sequencing_group_by_id(
            sequencing_group_id_transform_to_raw(mutation_result['id'])
        )
        self.assertEqual(
            api_result.meta,
            {'test': 'test'},
        )

    # endregion SEQUENCING GROUP TESTS



import logging
import tempfile


def setup_logger(
    dataset: str, name: str, level: str = 'INFO', log_file: str = None
) -> logging.Logger:
    """Set up logger for audit reviews."""
    logger = logging.getLogger(name)

    # Avoid duplicate handlers
    if logger.handlers:
        return logging.LoggerAdapter(logger, {'dataset': dataset})

    formatter = logging.Formatter(
        fmt='%(asctime)s %(levelname)s %(module)s:%(lineno)d - %(dataset)s :: %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
    )

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # Optional file handler for audit persistence
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    logger.setLevel(getattr(logging, level.upper()))
    logger.propagate = False

    return logging.LoggerAdapter(logger, {'dataset': dataset})


class BucketAuditLogger:
    """Logging wrapper for bucket audit operations."""

    def __init__(self, dataset: str, name: str):
        """Initialize the audit logger."""
        self.log_file = tempfile.NamedTemporaryFile(  # pylint: disable=consider-using-with
            delete=False
        ).name
        self.logger = setup_logger(dataset, name, log_file=self.log_file)

    def info(self, message: str):
        """Log an info message."""
        self.logger.info(message)

    def info_nl(self, message: str):
        """Log an info message and a newline."""
        self.logger.info(message)
        self.logger.info('')

    def warning(self, message: str):
        """Log a warning message."""
        self.logger.warning(message)

    def error(self, message: str):
        """Log an error message."""
        self.logger.error(message)



import unittest
from datetime import date, datetime

from api.utils.dates import (
    get_invoice_month_range,
    parse_date_only_string,
    reformat_datetime,
)


class TestApiUtils(unittest.TestCase):
    """Test API utils functions"""

    def test_parse_date_only_string(self):
        """
        Test parse_date_only_string function
        """
        result_none = parse_date_only_string(None)
        self.assertEqual(None, result_none)

        result_date = parse_date_only_string('2021-01-10')
        self.assertEqual(2021, result_date.year)
        self.assertEqual(1, result_date.month)
        self.assertEqual(10, result_date.day)

        # test exception
        invalid_date_str = '123456789'
        with self.assertRaises(ValueError) as context:
            parse_date_only_string(invalid_date_str)

        self.assertTrue(
            f'Date could not be converted: {invalid_date_str}' in str(context.exception)
        )

    def test_get_invoice_month_range(self):
        """
        Test get_invoice_month_range function
        """
        jan_2021 = datetime.strptime('2021-01-10', '%Y-%m-%d').date()
        res_jan_2021 = get_invoice_month_range(jan_2021)

        # there is 3 (INVOICE_DAY_DIFF) days difference between invoice month st and end
        self.assertEqual(
            (date(2020, 12, 29), date(2021, 2, 3)),
            res_jan_2021,
        )

        dec_2021 = datetime.strptime('2021-12-10', '%Y-%m-%d').date()
        res_dec_2021 = get_invoice_month_range(dec_2021)

        # there is 3 (INVOICE_DAY_DIFF) days difference between invoice month st and end
        self.assertEqual(
            (date(2021, 11, 28), date(2022, 1, 3)),
            res_dec_2021,
        )

    def test_reformat_datetime(self):
        """
        Test reformat_datetime function
        """
        in_format = '%Y-%m-%d'
        out_format = '%d/%m/%Y'

        result_none = reformat_datetime(None, in_format, out_format)
        self.assertEqual(None, result_none)

        result_formatted = reformat_datetime('2021-11-09', in_format, out_format)
        self.assertEqual('09/11/2021', result_formatted)

        # test exception
        invalid_date_str = '123456789'
        with self.assertRaises(ValueError) as context:
            reformat_datetime(invalid_date_str, in_format, out_format)

        self.assertTrue(
            f'Date could not be converted: {invalid_date_str}' in str(context.exception)
        )



from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

import api.routes.cohort
import metamist.models
from db.python.layers import SampleLayer
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from models.models.cohort import CohortBody, CohortCriteria, NewCohort
from models.utils.cohort_template_id_format import cohort_template_id_format
from models.utils.sequencing_group_id_format import sequencing_group_id_format
from scripts.create_custom_cohort import get_cohort_spec, main


class TestCohortBuilderBasic(DbIsolatedTest):
    """Test basic functionality for the custom cohort builder script"""

    @run_as_sync
    async def setUp(self):
        super().setUp()

    @run_as_sync
    async def mock_ccfc(self, project, body_create_cohort_from_criteria):
        """Mock by directly calling the API route"""
        self.assertEqual(project, self.project_name)
        return await api.routes.cohort.create_cohort_from_criteria(
            CohortBody(**body_create_cohort_from_criteria['cohort_spec'].to_dict()),
            CohortCriteria(
                **body_create_cohort_from_criteria['cohort_criteria'].to_dict()
            ),
            self.connection,
            body_create_cohort_from_criteria['dry_run'],
        )

    @run_as_sync
    async def test_get_cohort_spec(self):
        """Test get_cohort_spec(), invoked by the creator script"""
        ctemplate_id = cohort_template_id_format(28)
        result = get_cohort_spec('My cohort', 'Describing the cohort', ctemplate_id)
        self.assertIsInstance(result, metamist.models.CohortBody)
        self.assertEqual(result.name, 'My cohort')
        self.assertEqual(result.description, 'Describing the cohort')
        self.assertEqual(result.template_id, ctemplate_id)

    @run_as_sync
    @patch('metamist.apis.CohortApi.create_cohort_from_criteria')
    async def test_build_empty_cohort(self, mock):
        """Test main with an empty cohort"""
        mock.side_effect = self.mock_ccfc
        with self.assertRaises(ValueError) as context:
            _ = main(
                project=self.project_name,
                cohort_body_spec=metamist.models.CohortBody(
                    name='Empty cohort', description='No criteria'
                ),
                projects=[self.project_name],
                sg_ids_internal=[],
                excluded_sg_ids=[],
                sg_technologies=[],
                sg_platforms=[],
                sg_types=[],
                sample_types=[],
                dry_run=False,
            )
        mock.assert_called_once()
        self.assertIn(
            'criteria resulted in no sequencing groups',
            str(context.exception),
        )


def get_sample_model(
    eid, s_type='blood', sg_type='genome', tech='short-read', plat='illumina'
):
    """Create a minimal sample"""
    return SampleUpsertInternal(
        meta={},
        external_ids={PRIMARY_EXTERNAL_ORG: f'EXID{eid}'},
        type=s_type,
        sequencing_groups=[
            SequencingGroupUpsertInternal(
                type=sg_type,
                technology=tech,
                platform=plat,
                meta={},
                assays=[],
            ),
        ],
    )


class TestCohortBuilderData(DbIsolatedTest):
    """Test functionality for the custom cohort builder script that requires data"""

    @run_as_sync
    async def setUp(self):
        super().setUp()
        self.samplel = SampleLayer(self.connection)

        self.sA = await self.samplel.upsert_sample(
            get_sample_model('A', 'blood', 'genome', 'short-read', 'illumina')
        )
        self.sB = await self.samplel.upsert_sample(
            get_sample_model('B', 'blood', 'genome', 'short-read', 'illumina')
        )
        self.sC = await self.samplel.upsert_sample(
            get_sample_model('C', 'blood', 'genome', 'short-read', 'illumina')
        )

        self.sgA = sequencing_group_id_format(self.sA.sequencing_groups[0].id)
        self.sgB = sequencing_group_id_format(self.sB.sequencing_groups[0].id)
        self.sgC = sequencing_group_id_format(self.sC.sequencing_groups[0].id)

    @run_as_sync
    async def mock_ccfc(self, project, body_create_cohort_from_criteria):
        """Mock by directly calling the API route"""
        self.assertEqual(project, self.project_name)
        return await api.routes.cohort.create_cohort_from_criteria(
            CohortBody(**body_create_cohort_from_criteria['cohort_spec'].to_dict()),
            CohortCriteria(
                **body_create_cohort_from_criteria['cohort_criteria'].to_dict()
            ),
            self.connection,
            body_create_cohort_from_criteria['dry_run'],
        )

    @run_as_sync
    @patch('metamist.apis.CohortApi.create_cohort_from_criteria')
    async def test_empty_main(self, mock):
        """Test main with no criteria other than project"""
        mock.side_effect = self.mock_ccfc

        result = main(
            project=self.project_name,
            cohort_body_spec=metamist.models.CohortBody(
                name='Test cohort', description='Project criteria'
            ),
            projects=[self.project_name],
            sg_ids_internal=[],
            excluded_sg_ids=[],
            sg_technologies=[],
            sg_platforms=[],
            sg_types=[],
            sample_types=[],
            dry_run=False,
        )
        mock.assert_called_once()
        self.assertIsInstance(result, NewCohort)
        self.assertIsInstance(result.cohort_id, str)
        self.assertListEqual(
            result.sequencing_group_ids, [self.sgA, self.sgB, self.sgC]
        )
        self.assertEqual(result.dry_run, False)

    @run_as_sync
    @patch('metamist.apis.CohortApi.create_cohort_from_criteria')
    async def test_epic_main(self, mock):
        """Test main with every criteria"""
        mock.side_effect = self.mock_ccfc
        result = main(
            project=self.project_name,
            cohort_body_spec=metamist.models.CohortBody(
                name='Epic cohort', description='Every criterion'
            ),
            projects=[self.project_name],
            sg_ids_internal=[self.sgA],
            excluded_sg_ids=[self.sgB],
            sg_technologies=['short-read'],
            sg_platforms=['illumina'],
            sg_types=['genome'],
            sample_types=['blood'],
            dry_run=False,
        )
        mock.assert_called_once()
        self.assertEqual(mock.call_args.kwargs['project'], self.project_name)

        body = mock.call_args.kwargs['body_create_cohort_from_criteria']
        spec = body['cohort_spec']
        self.assertEqual(spec.name, 'Epic cohort')
        self.assertEqual(spec.description, 'Every criterion')

        criteria = body['cohort_criteria']
        self.assertIsInstance(criteria, metamist.models.CohortCriteria)
        self.assertListEqual(criteria.projects, [self.project_name])
        self.assertListEqual(criteria.sg_ids_internal, [self.sgA])
        self.assertListEqual(criteria.excluded_sgs_internal, [self.sgB])
        self.assertListEqual(criteria.sg_technology, ['short-read'])
        self.assertListEqual(criteria.sg_platform, ['illumina'])
        self.assertListEqual(criteria.sg_type, ['genome'])
        self.assertListEqual(criteria.sample_type, ['blood'])

        self.assertFalse(body['dry_run'])

        self.assertIsInstance(result, NewCohort)
        self.assertListEqual(result.sequencing_group_ids, [self.sgA])
        self.assertEqual(result.dry_run, False)



from typing import Any

from db.python.filters import GenericFilter
from db.python.layers.participant import ParticipantLayer
from db.python.tables.participant import ParticipantFilter
from models.base import PRIMARY_EXTERNAL_ORG
from models.models.assay import AssayUpsertInternal
from models.models.participant import ParticipantUpsertInternal
from models.models.sample import SampleUpsertInternal
from models.models.sequencing_group import SequencingGroupUpsertInternal
from test.testbase import DbIsolatedTest, run_as_sync


def get_participant_to_insert(id_suffix='1'):
    """Helper function to create a participant object for insertion into the database"""
    return ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'P0' + id_suffix},
        meta={'pmeta': 'pvalue'},
        reported_sex=2,
        reported_gender='FEMALE',
        karyotype='XX',
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'S0' + id_suffix},
                type='blood',
                meta={'smeta': 'svalue'},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        external_ids={'default': 'SG0' + id_suffix},
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A0' + id_suffix},
                                meta={
                                    'ameta': 'avalue',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            )
                        ],
                    )
                ],
            )
        ],
    )


class TestParticipant(DbIsolatedTest):
    """Test participant related functionality"""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()  # type: ignore

        self.player = ParticipantLayer(self.connection)

    @run_as_sync
    async def test_query_by_ids(self):
        """Test query"""

        p = await self.player.upsert_participant(get_participant_to_insert())

        ps = await self.player.query(ParticipantFilter(id=GenericFilter(eq=p.id)))

        self.assertEqual(len(ps), 1)
        self.assertEqual(ps[0].id, p.id)

        ps = await self.player.query(ParticipantFilter(id=GenericFilter(in_=[-1])))
        self.assertEqual(len(ps), 0)

    @run_as_sync
    async def test_query_by_exids(self):
        """Test query"""

        p = get_participant_to_insert()
        p.external_ids = {
            PRIMARY_EXTERNAL_ORG: 'P01',
            'external_org': 'ex01',
        }

        p = await self.player.upsert_participant(p)

        ps = await self.player.query(
            ParticipantFilter(external_id=GenericFilter(eq='P01'))
        )

        self.assertEqual(len(ps), 1)
        self.assertEqual(ps[0].id, p.id)

        ps = await self.player.query(
            ParticipantFilter(external_id=GenericFilter(in_=['ex01']))
        )
        self.assertEqual(len(ps), 1)
        self.assertEqual(ps[0].id, p.id)

        ps = await self.player.query(
            ParticipantFilter(external_id=GenericFilter(in_=['ex02']))
        )
        self.assertEqual(len(ps), 0)

    @run_as_sync
    async def test_graphql_query_by_id(self):
        """Test query by id using graphql"""
        p = await self.player.upsert_participant(get_participant_to_insert())

        q = """
query TestGraphqlQueryById($projectName: String!, $pid: Int!) {
    project(name: $projectName) {
        participants(id: { in_: [$pid] }) {
            id
        }
    }
}
"""
        resp = await self.run_graphql_query_async(
            q, {'projectName': self.project_name, 'pid': p.id}
        )
        assert resp is not None

        self.assertEqual(1, len(resp['project']['participants']))

    @run_as_sync
    async def test_query_by_sample(self):
        """Test query"""

        p = await self.player.upsert_participant(get_participant_to_insert())

        ps = await self.player.query(
            ParticipantFilter(
                sample=ParticipantFilter.ParticipantSampleFilter(
                    external_id=GenericFilter(in_=['S01'])
                )
            )
        )

        self.assertEqual(len(ps), 1)
        self.assertEqual(ps[0].id, p.id)

        ps = await self.player.query(
            ParticipantFilter(
                sample=ParticipantFilter.ParticipantSampleFilter(
                    external_id=GenericFilter(in_=['S01-NOT_PRESENT'])
                )
            )
        )
        self.assertEqual(len(ps), 0)

    @run_as_sync
    async def test_query_with_offset(self):
        """Test query providing an offset and a limit"""

        p1 = await self.player.upsert_participant(get_participant_to_insert('1'))
        p2 = await self.player.upsert_participant(get_participant_to_insert('2'))

        participants = await self.player.query(
            ParticipantFilter(project=GenericFilter(eq=self.project_id)),
            limit=1,
        )

        self.assertEqual(len(participants), 1)
        self.assertEqual(participants[0].id, p1.id)

        participants = await self.player.query(
            ParticipantFilter(project=GenericFilter(eq=self.project_id)),
            limit=1,
            skip=1,
        )

        self.assertEqual(len(participants), 1)
        self.assertEqual(participants[0].id, p2.id)

    @run_as_sync
    async def test_upsert_participant_with_phenotypes(self):
        """Test upserting participant with phenotypes"""

        phenotypes: dict[str, Any] = {
            'phenotype1': 'value1',
            'phenotype2': {'number': 123},
        }
        p = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
                meta={},
                samples=[],
                phenotypes=phenotypes,
            )
        )

        q = """
        query GetParticipant($pid: Int!) {
            participant(id: $pid) {
                id
                phenotypes
            }
        }"""

        resp = await self.run_graphql_query_async(q, {'pid': p.id})

        resp_participant = resp['participant']

        self.assertEqual(resp_participant['id'], p.id)

        self.assertDictEqual(resp_participant['phenotypes'], phenotypes)

    @run_as_sync
    async def test_upsert_participant_with_phenotypes_twice(self):
        """Test upserting and then updating participant with phenotypes"""

        phenotypes1: dict[str, Any] = {
            'phenotype1': 'value1',
            'phenotype2': {'number': 123},
        }

        phenotypes2: dict[str, Any] = {
            'phenotype1': 'value2',
            'phenotype2': {'number': 345},
            'phenotype3': {'number': 678},
        }

        p1 = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
                meta={},
                samples=[],
                phenotypes=phenotypes1,
            )
        )

        p2 = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                id=p1.id,
                meta={},
                samples=[],
                phenotypes=phenotypes2,
            )
        )

        # ensure second upsert didn't create a new participant
        self.assertEqual(p1.id, p2.id)

        q = """
        query GetParticipant($pid: Int!) {
            participant(id: $pid) {
                id
                phenotypes
            }
        }"""

        resp = await self.run_graphql_query_async(q, {'pid': p1.id})

        resp_participant = resp['participant']

        self.assertEqual(resp_participant['id'], p2.id)

        self.assertDictEqual(resp_participant['phenotypes'], phenotypes2)

    @run_as_sync
    async def test_graphql_upsert_participant_with_phenotypes(self):
        """Test upserting and then updating participant with phenotypes, via graphql"""

        phenotypes1: dict[str, Any] = {
            'phenotype1': 'value1',
            'phenotype2': {'number': 123},
        }

        phenotypes2: dict[str, Any] = {
            'phenotype1': 'value2',
            'phenotype2': {'number': 345},
            'phenotype3': {'number': 678},
        }

        mutation = """
        mutation ParticipantPhenotype($participants:[ParticipantUpsertInput!]!, $project: String!) {
            participant {
                upsertParticipants(project: $project, participants:$participants) {
                    id
                    phenotypes
                }
            }
        }
        """

        p1_resp = await self.run_graphql_query_async(
            mutation,
            {
                'project': self.project_name,
                'participants': [
                    {
                        'externalIds': {PRIMARY_EXTERNAL_ORG: 'Demeter'},
                        'phenotypes': phenotypes1,
                    }
                ],
            },
        )

        p1 = p1_resp['participant']['upsertParticipants'][0]

        p1_resp = await self.run_graphql_query_async(
            mutation,
            {
                'project': self.project_name,
                'participants': [
                    {
                        'id': p1['id'],
                        'phenotypes': phenotypes2,
                    }
                ],
            },
        )

        p2 = p1_resp['participant']['upsertParticipants'][0]

        # ensure second upsert didn't create a new participant
        self.assertEqual(p1['id'], p2['id'])

        q = """
        query GetParticipant($pid: Int!) {
            participant(id: $pid) {
                id
                phenotypes
            }
        }"""

        resp = await self.run_graphql_query_async(q, {'pid': p1['id']})

        resp_participant = resp['participant']

        self.assertEqual(resp_participant['id'], p2['id'])

        self.assertDictEqual(resp_participant['phenotypes'], phenotypes2)



from datetime import date, datetime
from unittest import mock

from db.python.filters import GenericFilter
from db.python.layers import AnalysisLayer, SampleLayer, SequencingGroupLayer
from db.python.tables.sequencing_group import SequencingGroupFilter
from models.enums.analysis import AnalysisStatus
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AnalysisInternal,
    AssayUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from test.testbase import DbIsolatedTest, run_as_sync


def get_sample_model():
    """
    Get sample model with sequencing-groups, return in a function
    to protect against any mutation to this model
    """
    return SampleUpsertInternal(
        meta={},
        external_ids={PRIMARY_EXTERNAL_ORG: 'EX_ID'},
        sequencing_groups=[
            SequencingGroupUpsertInternal(
                type='genome',
                technology='short-read',
                platform='ILLUMINA',
                meta={
                    'meta-key': 'meta-value',
                },
                external_ids={'ext': 'some-ext-id'},
                assays=[
                    AssayUpsertInternal(
                        type='sequencing',
                        external_ids={},
                        meta={
                            'sequencing_type': 'genome',
                            'sequencing_platform': 'short-read',
                            'sequencing_technology': 'illumina',
                        },
                    )
                ],
            )
        ],
    )


class TestSequencingGroup(DbIsolatedTest):
    """Test sequencing groups business logic"""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()
        self.sglayer = SequencingGroupLayer(self.connection)
        self.slayer = SampleLayer(self.connection)
        self.alayer = AnalysisLayer(self.connection)

    @run_as_sync
    async def test_empty_query(self):
        """
        Test empty IDs to see the query construction
        """
        sgs = await self.sglayer.query(SequencingGroupFilter(id=GenericFilter(in_=[])))
        self.assertEqual(len(sgs), 0)

    @run_as_sync
    async def test_insert_sequencing_group(self):
        """Test inserting and fetching a sequencing group"""
        sample_to_insert = get_sample_model()
        sample = await self.slayer.upsert_sample(sample_to_insert)
        sg_id = sample.sequencing_groups[0].id
        sg = await self.sglayer.get_sequencing_group_by_id(sg_id)

        inserted_sg = sample_to_insert.sequencing_groups[0]
        self.assertEqual(inserted_sg.id, sg_id)
        self.assertEqual(inserted_sg.type, sg.type)
        self.assertEqual(inserted_sg.technology, sg.technology)
        self.assertEqual(inserted_sg.platform.lower(), sg.platform.lower())
        self.assertDictEqual(inserted_sg.meta, sg.meta)

    @run_as_sync
    async def test_update_sequencing_group(self):
        """Test updating metadata on a sequencing group"""
        sample = await self.slayer.upsert_sample(get_sample_model())

        upsert_sg = SequencingGroupUpsertInternal(
            id=sample.sequencing_groups[0].id,
            meta={'another-meta': 'field'},
        )
        # Check that id is being returned when seqg created
        sg_return = await self.sglayer.upsert_sequencing_groups([upsert_sg])
        self.assertEqual(sg_return[0].id, upsert_sg.id)

        sg = await self.sglayer.get_sequencing_group_by_id(
            sample.sequencing_groups[0].id
        )

        self.assertDictEqual(
            {'another-meta': 'field', 'meta-key': 'meta-value'}, sg.meta
        )

    @run_as_sync
    async def test_auto_deprecation_of_old_sequencing_group(self):
        """Test creating a sequencing-group, and test the old one is archived"""
        sample = await self.slayer.upsert_sample(get_sample_model())

        # self.sglayer.get_sequencing_groups_by_ids()

        new_upsert = SampleUpsertInternal(
            id=sample.id,
            sequencing_groups=[
                SequencingGroupUpsertInternal(
                    type='genome',
                    technology='short-read',
                    platform='ILLUMINA',
                    meta={
                        'meta-key': 'meta-value',
                    },
                    external_ids={},
                    assays=[
                        # include an empty assay with ID to ensure it gets added to the sg
                        AssayUpsertInternal(
                            id=sample.sequencing_groups[0].assays[0].id,
                        ),
                        # new assay to trigger deprecation
                        AssayUpsertInternal(
                            type='sequencing',
                            external_ids={'second-key': 'second-sequencing-object'},
                            meta={
                                'sequencing_type': 'genome',
                                'sequencing_platform': 'short-read',
                                'sequencing_technology': 'illumina',
                            },
                        ),
                    ],
                )
            ],
        )

        updated_sample = await self.slayer.upsert_sample(new_upsert)

        old_sg = await self.sglayer.get_sequencing_group_by_id(
            sample.sequencing_groups[0].id
        )
        # now check the existing sequencing group was archived
        self.assertTrue(old_sg.archived)

        # check that the "active" sequencing group is the new one
        active_sgs = await self.sglayer.query(
            SequencingGroupFilter(
                sample=SequencingGroupFilter.SequencingGroupSampleFilter(
                    id=GenericFilter(eq=sample.id)
                )
            )
        )

        self.assertTrue(all(not sg.archived for sg in active_sgs))
        self.assertEqual(len(active_sgs), 1)
        self.assertEqual(updated_sample.sequencing_groups[0].id, active_sgs[0].id)

    @run_as_sync
    async def test_query_with_assay_metadata(self):
        """Test searching with an assay metadata filter"""
        sample_to_insert = get_sample_model()

        # Add extra sequencing group
        sample_to_insert.sequencing_groups.append(
            SequencingGroupUpsertInternal(
                type='exome',
                technology='short-read',
                platform='ILLUMINA',
                meta={
                    'meta-key': 'meta-value',
                },
                external_ids={},
                assays=[
                    AssayUpsertInternal(
                        type='sequencing',
                        external_ids={},
                        meta={
                            'sequencing_type': 'exome',
                            'sequencing_platform': 'short-read',
                            'sequencing_technology': 'illumina',
                        },
                    )
                ],
            )
        )

        # Create in database
        sample = await self.slayer.upsert_sample(sample_to_insert)

        # Query for genome assay metadata
        sgs = await self.sglayer.query(
            SequencingGroupFilter(
                assay=SequencingGroupFilter.SequencingGroupAssayFilter(
                    meta={'sequencing_type': GenericFilter(eq='genome')}
                )
            )
        )
        self.assertEqual(len(sgs), 1)
        self.assertEqual(sgs[0].id, sample.sequencing_groups[0].id)

        # Query for exome assay metadata
        sgs = await self.sglayer.query(
            SequencingGroupFilter(
                assay=SequencingGroupFilter.SequencingGroupAssayFilter(
                    meta={'sequencing_type': GenericFilter(eq='exome')}
                )
            )
        )
        self.assertEqual(len(sgs), 1)
        self.assertEqual(sgs[0].id, sample.sequencing_groups[1].id)

    @run_as_sync
    async def test_query_with_creation_date(self):
        """Test fetching using a creation date filter"""
        sample_to_insert = get_sample_model()
        await self.slayer.upsert_sample(sample_to_insert)

        # There's a race condition here -- don't run this near UTC midnight!
        today = datetime.utcnow().date()

        # Query for sequencing group with creation date before today
        sgs = await self.sglayer.query(
            SequencingGroupFilter(created_on=GenericFilter(lt=today))
        )
        self.assertEqual(len(sgs), 0)

        # Query for sequencing group with creation date today
        sgs = await self.sglayer.query(
            SequencingGroupFilter(created_on=GenericFilter(eq=today))
        )
        self.assertEqual(len(sgs), 1)

        sgs = await self.sglayer.query(
            SequencingGroupFilter(created_on=GenericFilter(lte=today))
        )
        self.assertEqual(len(sgs), 1)

        sgs = await self.sglayer.query(
            SequencingGroupFilter(created_on=GenericFilter(gte=today))
        )
        self.assertEqual(len(sgs), 1)

        # Query for sequencing group with creation date today
        sgs = await self.sglayer.query(
            SequencingGroupFilter(created_on=GenericFilter(gt=today))
        )
        self.assertEqual(len(sgs), 0)

    @run_as_sync
    async def test_query_finds_sgs_which_have_cram_analysis(self):
        """Test querying for sequencing groups which have a cram or gvcf analysis"""
        sample_to_insert = get_sample_model()

        # Add extra sequencing group
        sample_to_insert.sequencing_groups.append(
            SequencingGroupUpsertInternal(
                type='exome',
                technology='short-read',
                platform='ILLUMINA',
                meta={
                    'meta-key': 'meta-value',
                },
                external_ids={},
                assays=[
                    AssayUpsertInternal(
                        type='sequencing',
                        external_ids={},
                        meta={
                            'sequencing_type': 'exome',
                            'sequencing_platform': 'short-read',
                            'sequencing_technology': 'illumina',
                        },
                    )
                ],
            )
        )

        # Create in database
        sample = await self.slayer.upsert_sample(sample_to_insert)

        # Create analysis for cram and gvcf
        await self.alayer.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[sample.sequencing_groups[0].id],
                meta={},
            )
        )
        await self.alayer.create_analysis(
            AnalysisInternal(
                type='gvcf',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[sample.sequencing_groups[1].id],
                meta={},
            )
        )

        # Query for cram analysis
        sgs = await self.sglayer.query(SequencingGroupFilter(has_cram=True))
        self.assertEqual(len(sgs), 1)
        self.assertEqual(sgs[0].id, sample.sequencing_groups[0].id)

        # Query for gvcf analysis
        sgs = await self.sglayer.query(SequencingGroupFilter(has_gvcf=True))
        self.assertEqual(len(sgs), 1)
        self.assertEqual(sgs[0].id, sample.sequencing_groups[1].id)

        # Query for both cram AND gvcf analysis
        sgs = await self.sglayer.query(
            SequencingGroupFilter(has_gvcf=True, has_cram=True)
        )
        self.assertEqual(len(sgs), 0)

        # Add first SG to gvcf analysis
        await self.alayer.create_analysis(
            AnalysisInternal(
                type='gvcf',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[sample.sequencing_groups[0].id],
                meta={},
            )
        )

        # Query for both cram AND gvcf analysis now that first SG has gvcf analysis
        sgs = await self.sglayer.query(
            SequencingGroupFilter(has_gvcf=True, has_cram=True)
        )
        self.assertEqual(len(sgs), 1)
        self.assertEqual(sgs[0].id, sample.sequencing_groups[0].id)

    @run_as_sync
    async def test_archiving_sequencing_groups(self):
        """Check that sequencing groups can be archived from graphql"""
        sample_model = SampleUpsertInternal(
            meta={},
            external_ids={PRIMARY_EXTERNAL_ORG: 'EXID1'},
            type='blood',
            sequencing_groups=[
                SequencingGroupUpsertInternal(
                    type='genome',
                    technology='short-read',
                    platform='illumina',
                    meta={},
                    assays=[],
                ),
                SequencingGroupUpsertInternal(
                    type='genome',
                    technology='short-read',
                    platform='illumina',
                    meta={},
                    assays=[],
                ),
                SequencingGroupUpsertInternal(
                    type='exome',
                    technology='short-read',
                    platform='illumina',
                    meta={},
                    assays=[],
                ),
            ],
        )

        sample = await self.slayer.upsert_sample(sample_model)
        assert sample.sequencing_groups
        sg1 = sample.sequencing_groups[0].to_external().id
        sg2 = sample.sequencing_groups[1].to_external().id

        assert sg1, sg2

        archive_result = await self.run_graphql_query_async(
            """
            mutation ArchiveSeqGroups($ids: [String!]!) {
                sequencingGroup {
                    archiveSequencingGroups(sequencingGroupIds:$ids) {
                        id
                        archived
                    }
                }
            }
            """,
            {'ids': [sg1, sg2]},
        )

        archived_sgs = archive_result['sequencingGroup']['archiveSequencingGroups']

        self.assertEqual(len(archived_sgs), 2)
        self.assertEqual(archived_sgs[0]['id'], sg1)
        self.assertEqual(archived_sgs[0]['archived'], True)
        self.assertEqual(archived_sgs[1]['id'], sg2)
        self.assertEqual(archived_sgs[1]['archived'], True)

    @run_as_sync
    async def test_history_no_sum(self):
        """Test the trivial case where there are no sequencing groups."""

        # Set up mocking for rows returned from the table query.
        with mock.patch(
            'db.python.connect.databases.Database.fetch_all', return_value=[]
        ):
            sg_table = self.sglayer.seqgt
            result = await sg_table.get_sequencing_group_counts_by_month([])

        self.assertDictEqual(result, {})

    @run_as_sync
    @mock.patch('db.python.tables.sequencing_group.date', wraps=date)
    async def test_history_sum_multiple_projects(self, mock_date):
        """Test the case where type:technology combinations are summed and held for the same project."""
        # Mock today's date.
        mock_date.today.return_value = date(year=2025, month=12, day=31)

        # Set up mocking for rows returned from the table query.
        rows_mock = [
            {
                'project': 0,
                'type': 'genome',
                'technology': 'short-read',
                'sg_date': date(2025, 10, 1),
                'num_sg': 2,
            },
            {
                'project': 0,
                'type': 'genome',
                'technology': 'short-read',
                'sg_date': date(2025, 11, 1),
                'num_sg': 4,
            },
            {
                'project': 1,
                'type': 'genome',
                'technology': 'short-read',
                'sg_date': date(2025, 10, 1),
                'num_sg': 3,
            },
            {
                'project': 1,
                'type': 'genome',
                'technology': 'short-read',
                'sg_date': date(2025, 11, 1),
                'num_sg': 5,
            },
        ]
        with mock.patch(
            'db.python.connect.databases.Database.fetch_all', return_value=rows_mock
        ):
            sg_table = self.sglayer.seqgt
            result = await sg_table.get_sequencing_group_counts_by_month([0])

        self.assertDictEqual(
            result,
            {
                0: {
                    date(2025, 10, 1): {
                        'genome|||short-read': 2,
                    },
                    date(2025, 11, 1): {
                        'genome|||short-read': 6,
                    },
                    date(2025, 12, 1): {
                        'genome|||short-read': 6,
                    },
                },
                1: {
                    date(2025, 10, 1): {
                        'genome|||short-read': 3,
                    },
                    date(2025, 11, 1): {
                        'genome|||short-read': 8,
                    },
                    date(2025, 12, 1): {
                        'genome|||short-read': 8,
                    },
                },
            },
        )

    @run_as_sync
    @mock.patch('db.python.tables.sequencing_group.date', wraps=date)
    async def test_history_partial_sum(self, mock_date):
        """Test the case where less types are present initially and more are added over time."""
        # Mock today's date.
        mock_date.today.return_value = date(year=2025, month=12, day=31)

        # Set up mocking for rows returned from the table query.
        rows_mock = [
            {
                'project': 0,
                'type': 'genome',
                'technology': 'short-read',
                'sg_date': date(2025, 10, 1),
                'num_sg': 2,
            },
            {
                'project': 0,
                'type': 'genome',
                'technology': 'long-read',
                'sg_date': date(2025, 11, 1),
                'num_sg': 3,
            },
            {
                'project': 0,
                'type': 'chip',
                'technology': 'short-read',
                'sg_date': date(2025, 10, 1),
                'num_sg': 4,
            },
            {
                'project': 0,
                'type': 'chip',
                'technology': 'long-read',
                'sg_date': date(2025, 11, 1),
                'num_sg': 5,
            },
        ]
        with mock.patch(
            'db.python.connect.databases.Database.fetch_all', return_value=rows_mock
        ):
            sg_table = self.sglayer.seqgt
            result = await sg_table.get_sequencing_group_counts_by_month([0])

        self.assertDictEqual(
            result,
            {
                0: {
                    date(2025, 10, 1): {
                        'genome|||short-read': 2,
                        'chip|||short-read': 4,
                    },
                    date(2025, 11, 1): {
                        'genome|||short-read': 2,
                        'genome|||long-read': 3,
                        'chip|||short-read': 4,
                        'chip|||long-read': 5,
                    },
                    date(2025, 12, 1): {
                        'genome|||short-read': 2,
                        'genome|||long-read': 3,
                        'chip|||short-read': 4,
                        'chip|||long-read': 5,
                    },
                }
            },
        )



# pylint: disable=protected-access
import datetime
from test.testbase import run_as_sync
from test.testbqbase import BqTest
from textwrap import dedent
from typing import Any
from unittest import mock

import google.cloud.bigquery as bq

from db.python.tables.bq.billing_filter import BillingFilter
from db.python.tables.bq.billing_gcp_daily import BillingGcpDailyTable
from db.python.tables.bq.generic_bq_filter import GenericBQFilter
from db.python.utils import InternalError
from models.models import BillingColumn, BillingTotalCostQueryModel


class TestBillingGcpDailyTable(BqTest):
    """Test BillingRawTable and its methods"""

    def setUp(self):
        super().setUp()

        # setup table object
        self.table_obj = BillingGcpDailyTable(self.connection)

    def test_query_to_partitioned_filter(self):
        """Test query to partitioned filter conversion"""

        # given
        start_date = '2023-01-01'
        end_date = '2024-01-01'
        filters: dict[BillingColumn, str | list[Any] | dict[Any, Any]] = {
            BillingColumn.TOPIC: 'TEST_TOPIC'
        }

        # expected
        expected_filter = BillingFilter(
            day=GenericBQFilter(
                gte=datetime.datetime(2023, 1, 1, 0, 0),
                lte=datetime.datetime(2024, 1, 1, 0, 0),
            ),
            part_time=GenericBQFilter(
                gte=datetime.datetime(2023, 1, 1, 0, 0),
                lte=datetime.datetime(2024, 1, 8, 0, 0),  # 7 days added
            ),
            topic=GenericBQFilter(eq='TEST_TOPIC'),
        )

        query = BillingTotalCostQueryModel(
            fields=[],  # not relevant for this test, but can't be null generally
            start_date=start_date,
            end_date=end_date,
            filters=filters,
        )
        filter_ = BillingGcpDailyTable._query_to_partitioned_filter(query)

        # BillingFilter has __eq__ method, so we can compare them directly
        self.assertEqual(expected_filter, filter_)

    def test_error_no_connection(self):
        """Test No connection exception"""

        with self.assertRaises(InternalError) as context:
            BillingGcpDailyTable(None)

        self.assertTrue(
            "No connection was provided to the table 'BillingGcpDailyTable'"
            in str(context.exception)
        )

    def test_get_table_name(self):
        """Test get_table_name"""

        # table name is set in the class
        given_table_name = 'TEST_TABLE_NAME'

        # set table name
        self.table_obj.table_name = given_table_name

        # test get table name function
        table_name = self.table_obj.get_table_name()

        self.assertEqual(given_table_name, table_name)

    @run_as_sync
    async def test_last_loaded_day_return_valid_day(self):
        """Test _last_loaded_day"""

        given_last_day = '2021-01-01 00:00:00'

        # mock BigQuery result

        self.bq_result.result.return_value = [
            mock.MagicMock(
                spec=bq.Row,
                last_loaded_day=datetime.datetime.strptime(
                    given_last_day, '%Y-%m-%d %H:%M:%S'
                ),
            )
        ]  # 2021-01-01

        # test get table name function
        last_loaded_day = await self.table_obj._last_loaded_day()

        self.assertEqual(given_last_day, last_loaded_day)

    @run_as_sync
    async def test_last_loaded_day_return_none(self):
        """Test _last_loaded_day as None"""

        # mock BigQuery result as empty list
        self.bq_result.result.return_value = []

        # test get table name function
        last_loaded_day = await self.table_obj._last_loaded_day()

        self.assertEqual(None, last_loaded_day)

    def test_prepare_daily_cost_subquery(self):
        """Test _prepare_daily_cost_subquery"""

        self.table_obj.table_name = 'TEST_TABLE_NAME'

        # given
        given_field = BillingColumn.COST
        given_query_params: list[Any] = []
        given_last_loaded_day = '2021-01-01 00:00:00'

        (
            query_params,
            daily_cost_field,
            daily_cost_join,
        ) = self.table_obj._prepare_daily_cost_subquery(
            given_field, given_query_params, given_last_loaded_day
        )

        # expected
        expected_daily_cost_join = """LEFT JOIN (
            SELECT
                cost as field,
                cost_category,
                SUM(cost) as cost
            FROM
            `TEST_TABLE_NAME`
            WHERE day = TIMESTAMP(@last_loaded_day)

        AND part_time >= TIMESTAMP(@last_loaded_day)
        AND part_time <= TIMESTAMP_ADD(
            TIMESTAMP(@last_loaded_day), INTERVAL 7 DAY
        )

            GROUP BY
                field,
                cost_category
        ) day
        ON month.field = day.field
        AND month.cost_category = day.cost_category
        """

        self.assertEqual(
            [
                bq.ScalarQueryParameter(
                    'last_loaded_day', 'STRING', '2021-01-01 00:00:00'
                )
            ],
            query_params,
        )
        self.assertEqual(', day.cost as daily_cost', daily_cost_field)
        self.assertEqual(dedent(expected_daily_cost_join), dedent(daily_cost_join))

    @run_as_sync
    async def test_get_gcp_projects_return_empty_list(self):
        """Test get_gcp_projects as empty list"""

        # mock BigQuery result as empty list
        self.bq_result.result.return_value = []

        # test get table name function
        gcp_projects = await self.table_obj.get_gcp_projects()

        self.assertEqual([], gcp_projects)

    @run_as_sync
    async def test_get_gcp_projects_return_valid_list(self):
        """Test get_gcp_projects as empty list"""

        # mock BigQuery result as list of 2 records
        self.bq_result.result.return_value = [
            {'gcp_project': 'PROJECT1'},
            {'gcp_project': 'PROJECT2'},
        ]

        # test get table name function
        gcp_projects = await self.table_obj.get_gcp_projects()

        self.assertEqual(['PROJECT1', 'PROJECT2'], gcp_projects)



from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers import (
    AssayLayer,
    ParticipantLayer,
    ProjectInsightsLayer,
    SampleLayer,
)
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AssayUpsertInternal,
    ParticipantUpsertInternal,
    ProjectInsightsSummaryInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)

default_assay_meta = {
    'sequencing_type': 'genome',
    'sequencing_technology': 'short-read',
    'sequencing_platform': 'illumina',
}


def get_test_participant():
    """Do it like this to avoid an upsert writing the test value"""
    return ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id001'},
                meta={},
                type='blood',
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    'batch': 'M001',
                                    **default_assay_meta,
                                },
                            ),
                        ],
                    )
                ],
            )
        ],
    )


class TestProjectInsights(DbIsolatedTest):
    """Test project insights class containing project insights endpoints"""

    maxDiff = None

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()
        self.partl = ParticipantLayer(self.connection)
        self.pil = ProjectInsightsLayer(self.connection)
        self.sampl = SampleLayer(self.connection)
        self.seql = AssayLayer(self.connection)

    @run_as_sync
    async def test_project_insights_summary(self):
        """Test getting the summaries for all available projects"""

        await self.partl.upsert_participant(get_test_participant())

        result = await self.pil.get_project_insights_summary(
            project_names=[self.project_name], sequencing_types=['genome', 'exome']
        )

        expected = [
            ProjectInsightsSummaryInternal(
                project=self.project_id,
                dataset=self.project_name,  # for ProjectInsights, dataset is project.name
                sequencing_type='genome',
                sequencing_technology='short-read',
                total_families=0,
                total_participants=1,
                total_samples=1,
                total_sequencing_groups=1,
                total_crams=0,
                latest_annotate_dataset=None,
                latest_snv_es_index=None,
                latest_sv_es_index=None,
            ),
        ]

        self.assertEqual(result, expected)

    @run_as_sync
    async def test_project_insights_details(self):
        """Test getting the details for all available projects"""

        await self.partl.upsert_participant(get_test_participant())

        # There's not enough data set up to usefully verify the result
        _ = await self.pil.get_project_insights_details(
            project_names=[self.project_name], sequencing_types=['genome', 'exome']
        )



# pylint: disable=missing-function-docstring, too-many-public-methods, too-many-locals, too-many-lines
from typing import Any

from db.python.layers.assay import AssayLayer
from db.python.layers.family import FamilyLayer
from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from db.python.layers.sequencing_group import SequencingGroupLayer
from db.python.tables.project import ProjectPermissionsTable
from models.models import PRIMARY_EXTERNAL_ORG, SampleUpsertInternal
from models.models.assay import AssayUpsertInternal
from models.models.participant import ParticipantUpsertInternal
from models.models.sequencing_group import SequencingGroupUpsertInternal
from test.test_participant import get_participant_to_insert
from test.testbase import DbIsolatedTest, run_as_sync

# @TODO would be good to add permissions testing to this, but first need a better
# way of mocking the test user and their permissions to make that possible

COMMENT_VERSION_FRAGMENT = """
    fragment CommentVersionFragment on GraphQLCommentVersion {
        content
        author
        status
        timestamp
    }
"""

COMMENT_FRAGMENT = """
    fragment CommentFragment on GraphQLComment {
        id
        parentId
        author
        content
        createdAt
        updatedAt
        status

        entity {
            ... CommentEntityFragment
        }

        versions {
            ... CommentVersionFragment
        }
    }
"""

COMMENT_ENTITY_FRAGMENT = """
    fragment CommentEntityFragment on GraphQLCommentEntity {
        __typename
        ... on GraphQLProject {
            projectId: id
            projectName: name
        }
        ... on GraphQLSample {
            sampleId: id
            sampleExternalId: externalId
        }
        ... on GraphQLSequencingGroup {
            sequencingGroupId: id
            sample {
                id
            }
        }
        ... on GraphQLAssay {
            assayId: id
            sample {
                id
                project {
                    name
                }
            }
        }
        ... on GraphQLParticipant {
            participantId:id
            participantExternalId: externalId
        }
        ... on GraphQLFamily {
            familyId: id
            familyExternalId: externalId
        }
    }
"""

DISCUSSION_FRAGMENT = """
    fragment DiscussionFragment on GraphQLDiscussion {
        directComments {
            ... CommentFragment

            thread {
                ... CommentFragment
            }
        }

        relatedComments {
            ... CommentFragment

            thread {
                ... CommentFragment
            }
        }
    }
"""


ADD_COMMENT_TO_THREAD = """
    mutation AddCommentToThread($parentId: Int!, $content: String!) {
        comment {
            addCommentToThread(parentId: $parentId, content: $content) {
                ...CommentFragment
            }
        }
    }
"""

UPDATE_COMMENT = """
    mutation UpdateComment($id: Int!, $content: String!) {
        comment {
            updateComment(id: $id, content: $content) {
                ...CommentFragment
                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""

DELETE_COMMENT = """
    mutation DeleteComment($id: Int!) {
        comment {
            deleteComment(id: $id) {
                ...CommentFragment
                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""

RESTORE_COMMENT = """
    mutation RestoreComment($id: Int!) {
        comment {
            restoreComment(id: $id) {
                ...CommentFragment
                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


SAMPLE_COMMENTS = """
    query SampleComments($sampleId: String!) {
        sample(id:{eq: $sampleId}) {
            id
            discussion {
                ... DiscussionFragment
            }
        }
    }
"""

SAMPLE_ADD_COMMENT = """
    mutation AddCommentToSample($id: String!, $content: String!) {
        sample {
            addComment(id: $id, content: $content) {
                ...CommentFragment
                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


PROJECT_COMMENTS = """
    query ProjectComments($projectName: String!) {
        project(name: $projectName) {
            id
            name
            discussion {
                ... DiscussionFragment
            }
        }
    }
"""

PROJECT_ADD_COMMENT = """
    mutation AddCommentToProject($id: Int!, $content: String!) {
        project {
            addComment(id: $id, content: $content) {
                ...CommentFragment

                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


ASSAY_COMMENTS = """
    query AssayComments($assayId: Int!) {
        assay(id: $assayId) {
            id
            discussion {
                ... DiscussionFragment
            }
        }
    }
"""

ASSAY_ADD_COMMENT = """
    mutation AddCommentToAssay($id: Int!, $content: String!) {
        assay {
            addComment(id: $id, content: $content) {
                ...CommentFragment

                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


PARTICIPANT_COMMENTS = """
    query ParticipantComments($participantId: Int!) {
        participant(id: $participantId) {
            id
            discussion {
                ... DiscussionFragment
            }
        }
    }
"""

PARTICIPANT_ADD_COMMENT = """
    mutation AddCommentToParticipant($id: Int!, $content: String!) {
        participant {
            addComment(id: $id, content: $content) {
                ...CommentFragment

                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


FAMILY_COMMENTS = """
    query FamilyComments($familyId: Int!) {
        family(familyId: $familyId) {
            id
            discussion {
                ... DiscussionFragment
            }
        }
    }
"""

FAMILY_ADD_COMMENT = """
    mutation AddCommentToFamily($id: Int!, $content: String!) {
        family {
            addComment(id: $id, content: $content) {
                ...CommentFragment

                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


SEQUENCING_GROUP_COMMENTS = """
    query SequencingGroupComments($sequencingGroupId: String!) {
        sequencingGroups(id:{eq: $sequencingGroupId}) {
            id
            discussion {
                ... DiscussionFragment
            }
        }
    }
"""

SEQUENCING_GROUP_ADD_COMMENT = """
    mutation AddCommentToSequencingGroup($id: String!, $content: String!) {
        sequencingGroup {
            addComment(id: $id, content: $content) {
                ...CommentFragment

                thread {
                    ...CommentFragment
                }
            }
        }
    }
"""


COMMENT_RESULT_FRAGMENTS = f"""
    {COMMENT_VERSION_FRAGMENT}
    {COMMENT_ENTITY_FRAGMENT}
    {COMMENT_FRAGMENT}
"""

DISCUSSION_RESULT_FRAGMENTS = f"""
    {COMMENT_VERSION_FRAGMENT}
    {COMMENT_ENTITY_FRAGMENT}
    {COMMENT_FRAGMENT}
    {DISCUSSION_FRAGMENT}
"""

default_sequencing_meta = {
    'sequencing_type': 'genome',
    'sequencing_platform': 'short-read',
    'sequencing_technology': 'illumina',
}


def comment_uniq(comment: dict[str, Any]):
    return (comment['id'], comment['entity']['__typename'])


class TestComment(DbIsolatedTest):
    """Test commenting functionality"""

    async def add_comment_to_entity(
        self, comment_query: str, entity_id: str | int, content: str
    ) -> dict[str, Any]:
        create_comment_query = f"""
            {COMMENT_RESULT_FRAGMENTS}
            {comment_query}
        """

        create_comment_variables = {
            'id': entity_id,
            'content': content,
        }

        created_comment = await self.run_graphql_query_async(
            create_comment_query, variables=create_comment_variables
        )
        return created_comment

    async def get_discussion_on_entity(
        self, comment_query: str, variables: dict[str, int | str]
    ) -> dict[str, Any]:
        get_discussion_query = f"""
            {DISCUSSION_RESULT_FRAGMENTS}
            {comment_query}
        """

        requested_comment = await self.run_graphql_query_async(
            get_discussion_query, variables=variables
        )
        return requested_comment

    async def add_comment_to_sample(self, sample_id: str, comment: str):
        result = await self.add_comment_to_entity(
            SAMPLE_ADD_COMMENT, sample_id, comment
        )
        return result['sample']['addComment']

    async def get_discussion_on_sample(self, sample_id: str):
        result = await self.get_discussion_on_entity(
            SAMPLE_COMMENTS, {'sampleId': sample_id}
        )
        return result['sample'][0]['discussion']

    async def add_comment_to_project(self, project_id: int, comment: str):
        result = await self.add_comment_to_entity(
            PROJECT_ADD_COMMENT, project_id, comment
        )
        return result['project']['addComment']

    async def get_discussion_on_project(self, project_name: str):
        result = await self.get_discussion_on_entity(
            PROJECT_COMMENTS, {'projectName': project_name}
        )
        return result['project']['discussion']

    async def add_comment_to_assay(self, assay_id: int, comment: str):
        result = await self.add_comment_to_entity(ASSAY_ADD_COMMENT, assay_id, comment)
        return result['assay']['addComment']

    async def get_discussion_on_assay(self, assay_id: int):
        result = await self.get_discussion_on_entity(
            ASSAY_COMMENTS, {'assayId': assay_id}
        )
        return result['assay']['discussion']

    async def add_comment_to_participant(self, participant_id: int, comment: str):
        result = await self.add_comment_to_entity(
            PARTICIPANT_ADD_COMMENT, participant_id, comment
        )
        return result['participant']['addComment']

    async def get_discussion_on_participant(self, participant_id: int):
        result = await self.get_discussion_on_entity(
            PARTICIPANT_COMMENTS, {'participantId': participant_id}
        )
        return result['participant']['discussion']

    async def add_comment_to_family(self, family_id: int, comment: str):
        result = await self.add_comment_to_entity(
            FAMILY_ADD_COMMENT, family_id, comment
        )
        return result['family']['addComment']

    async def get_discussion_on_family(self, family_id: int):
        result = await self.get_discussion_on_entity(
            FAMILY_COMMENTS, {'familyId': family_id}
        )
        return result['family']['discussion']

    async def add_comment_to_sequencing_group(
        self, sequencing_group_id: str, comment: str
    ):
        result = await self.add_comment_to_entity(
            SEQUENCING_GROUP_ADD_COMMENT, sequencing_group_id, comment
        )
        return result['sequencingGroup']['addComment']

    async def get_discussion_on_sequencing_group(self, sequencing_group_id: int):
        result = await self.get_discussion_on_entity(
            SEQUENCING_GROUP_COMMENTS, {'sequencingGroupId': sequencing_group_id}
        )
        return result['sequencingGroups'][0]['discussion']

    async def add_comment_to_thread(
        self, parent_id: int, content: str
    ) -> dict[str, Any]:
        add_comment_to_thread_query = f"""
            {COMMENT_RESULT_FRAGMENTS}
            {ADD_COMMENT_TO_THREAD}
        """

        add_comment_to_thread_variables = {
            'parentId': parent_id,
            'content': content,
        }

        child_comment_result = await self.run_graphql_query_async(
            add_comment_to_thread_query, variables=add_comment_to_thread_variables
        )
        return child_comment_result['comment']['addCommentToThread']

    async def update_comment(self, id_: int, content: str) -> dict[str, Any]:
        update_comment_query = f"""
            {COMMENT_RESULT_FRAGMENTS}
            {UPDATE_COMMENT}
        """

        update_comment_variables = {
            'id': id_,
            'content': content,
        }

        update_comment_result = await self.run_graphql_query_async(
            update_comment_query, variables=update_comment_variables
        )
        return update_comment_result['comment']['updateComment']

    async def delete_comment(self, id_: int) -> dict[str, Any]:
        delete_comment_query = f"""
            {COMMENT_RESULT_FRAGMENTS}
            {DELETE_COMMENT}
        """

        delete_comment_variables = {'id': id_}

        delete_comment_result = await self.run_graphql_query_async(
            delete_comment_query, variables=delete_comment_variables
        )
        return delete_comment_result['comment']['deleteComment']

    async def restore_comment(self, id_: int) -> dict[str, Any]:
        restore_comment_query = f"""
            {COMMENT_RESULT_FRAGMENTS}
            {RESTORE_COMMENT}
        """

        restore_comment_variables = {'id': id_}

        restore_comment_result = await self.run_graphql_query_async(
            restore_comment_query, variables=restore_comment_variables
        )
        return restore_comment_result['comment']['restoreComment']

    # tests run in 'sorted by ascii' order
    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.slayer = SampleLayer(self.connection)
        self.assaylayer = AssayLayer(self.connection)
        self.player = ParticipantLayer(self.connection)
        self.flayer = FamilyLayer(self.connection)
        self.sglayer = SequencingGroupLayer(self.connection)

    @run_as_sync
    async def test_add_comment_to_sample(self):
        """Test adding a comment to a sample"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        sample_external = sample.to_external()

        comment_text = 'Sample Test Comment 1234'
        created_comment = await self.add_comment_to_sample(
            sample_external.id, comment_text
        )

        # Ensure that created comment has expected data
        self.assertEqual(created_comment['content'], comment_text)
        self.assertEqual(created_comment['status'], 'active')
        self.assertEqual(created_comment['entity']['sampleId'], sample_external.id)

        discussion = await self.get_discussion_on_sample(sample_external.id)

        # Ensure that comment can be requested after creating it, and that it is the same
        self.assertEqual(created_comment, discussion['directComments'][0])

    @run_as_sync
    async def test_add_comment_to_project(self):
        """Test adding a comment to a project"""

        project = self.connection.all_projects()[0]

        comment_text = 'Project Test Comment 1234'

        created_comment = await self.add_comment_to_project(project.id, comment_text)

        self.assertEqual(created_comment['content'], comment_text)
        self.assertEqual(created_comment['status'], 'active')
        self.assertEqual(created_comment['entity']['projectId'], project.id)

        requested_comment = await self.get_discussion_on_project(project.name)

        # Ensure that comment can be requested after creating it, and that it is the same
        self.assertEqual(
            created_comment,
            requested_comment['directComments'][0],
        )

    @run_as_sync
    async def test_add_comment_to_assay(self):
        """Test adding a comment to an assay"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        assay = await self.assaylayer.upsert_assay(
            assay=AssayUpsertInternal(
                sample_id=sample.id, meta=default_sequencing_meta, type='sequencing'
            )
        )
        assert assay.id

        comment_text = 'Assay Test Comment 1234'

        created_comment = await self.add_comment_to_assay(assay.id, comment_text)

        self.assertEqual(created_comment['content'], comment_text)
        self.assertEqual(created_comment['status'], 'active')
        self.assertEqual(created_comment['entity']['assayId'], assay.id)

        requested_comment = await self.get_discussion_on_assay(assay.id)

        # Ensure that comment can be requested after creating it, and that it is the same
        self.assertEqual(created_comment, requested_comment['directComments'][0])

    @run_as_sync
    async def test_add_comment_to_participant(self):
        """Test adding a comment to an participant"""

        participant = await self.player.upsert_participant(
            participant=ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'p_external_id'},
                meta={'pmeta': 'pvalue'},
                reported_sex=2,
                reported_gender='FEMALE',
                karyotype='XX',
            )
        )
        assert participant.id

        comment_text = 'Participant Test Comment 1234'

        created_comment = await self.add_comment_to_participant(
            participant.id, comment_text
        )

        self.assertEqual(created_comment['content'], comment_text)
        self.assertEqual(created_comment['status'], 'active')
        self.assertEqual(created_comment['entity']['participantId'], participant.id)

        requested_comment = await self.get_discussion_on_participant(participant.id)

        # Ensure that comment can be requested after creating it, and that it is the same
        self.assertEqual(created_comment, requested_comment['directComments'][0])

    @run_as_sync
    async def test_add_comment_to_family(self):
        """Test adding a comment to an family"""

        family = await self.flayer.create_family(
            external_ids={PRIMARY_EXTERNAL_ORG: 'f_external_id'},
        )

        comment_text = 'Family Test Comment 1234'
        created_comment = await self.add_comment_to_family(family, comment_text)

        self.assertEqual(created_comment['content'], comment_text)
        self.assertEqual(created_comment['status'], 'active')
        self.assertEqual(created_comment['entity']['familyId'], family)

        requested_comment = await self.get_discussion_on_family(family)

        # Ensure that comment can be requested after creating it, and that it is the same
        self.assertEqual(created_comment, requested_comment['directComments'][0])

    @run_as_sync
    async def test_add_comment_to_sequencing_group(self):
        """Test adding a comment to an sequencing_group"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        sequencing_groups = await self.sglayer.upsert_sequencing_groups(
            sequencing_groups=[
                SequencingGroupUpsertInternal(
                    sample_id=sample.id,
                    type='mtseq',
                    technology='short-read',
                    platform='oxford-nanopore',
                    assays=[
                        AssayUpsertInternal(
                            sample_id=sample.id,
                            meta=default_sequencing_meta,
                            type='sequencing',
                        )
                    ],
                )
            ]
        )
        sequencing_group = sequencing_groups[0]
        sequencing_group_id = sequencing_group.to_external().id
        assert sequencing_group_id

        comment_text = 'SequencingGroup Test Comment 1234'

        created_comment = await self.add_comment_to_sequencing_group(
            sequencing_group_id, comment_text
        )

        self.assertEqual(created_comment['content'], comment_text)
        self.assertEqual(created_comment['status'], 'active')
        self.assertEqual(
            created_comment['entity']['sequencingGroupId'],
            sequencing_group_id,
        )

        requested_comment = await self.get_discussion_on_sequencing_group(
            sequencing_group_id
        )

        # Ensure that comment can be requested after creating it, and that it is the same
        self.assertEqual(created_comment, requested_comment['directComments'][0])

    @run_as_sync
    async def test_add_comment_to_thread(self):
        """Test adding a comment to a thread"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        sample_external = sample.to_external()
        comment_text = 'Sample parent comment'

        parent_comment = await self.add_comment_to_sample(
            sample_external.id, comment_text
        )
        parent_id = parent_comment['id']

        child_comment = await self.add_comment_to_thread(
            parent_id=parent_id, content='Child comment in thread'
        )

        # Ensure that the child comment has the correct parent ID
        self.assertEqual(child_comment['parentId'], parent_id)

        requested_comment = await self.get_discussion_on_sample(sample_external.id)

        parent_comment_with_thread = requested_comment['directComments'][0]
        self.assertEqual(len(parent_comment_with_thread['thread']), 1)
        self.assertEqual(parent_comment_with_thread['thread'][0], child_comment)

    @run_as_sync
    async def test_updating_comment(self):
        """Test updating an existing parent comment"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        sample_external = sample.to_external()
        initial_parent_comment_text = 'Sample parent comment'

        parent_comment = await self.add_comment_to_sample(
            sample_external.id, initial_parent_comment_text
        )

        initial_child_comment_text = 'Sample child comment'

        child_comment = await self.add_comment_to_thread(
            parent_id=parent_comment['id'], content=initial_child_comment_text
        )

        updated_parent_comment_text = 'Updated Parent Comment Text'
        updated_child_comment_text = 'Updated Child Comment Text'

        # Update child comment first so that we can check that parent comment update
        # contains the updated child comment
        updated_child_comment = await self.update_comment(
            child_comment['id'], updated_child_comment_text
        )

        updated_parent_comment = await self.update_comment(
            parent_comment['id'], updated_parent_comment_text
        )

        sample_discussion = await self.get_discussion_on_sample(sample_external.id)

        self.assertEqual(updated_child_comment['content'], updated_child_comment_text)
        self.assertEqual(updated_parent_comment['content'], updated_parent_comment_text)
        self.assertEqual(
            updated_parent_comment['versions'][0]['content'],
            initial_parent_comment_text,
        )
        self.assertEqual(len(updated_parent_comment['versions']), 1)
        self.assertEqual(len(updated_child_comment['versions']), 1)
        self.assertEqual(
            updated_child_comment['versions'][0]['content'], initial_child_comment_text
        )
        self.assertEqual(
            # The child in the updated parent comment thread won't have a thread of its
            # own, so add it in here, but they otherwise should be identical
            updated_parent_comment['thread'][0] | {'thread': []},
            updated_child_comment,
        )
        self.assertEqual(sample_discussion['directComments'][0], updated_parent_comment)

    @run_as_sync
    async def test_deleting_and_restoring_comment(self):
        """Test deleting a comment and then restoring it"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        sample_external = sample.to_external()
        comment_text = 'Sample parent comment'

        created_comment = await self.add_comment_to_sample(
            sample_external.id, comment_text
        )
        self.assertEqual(created_comment['status'], 'active')

        deleted_comment = await self.delete_comment(created_comment['id'])

        self.assertEqual(deleted_comment['status'], 'deleted')
        self.assertEqual(len(deleted_comment['versions']), 1)

        restored_comment = await self.restore_comment(created_comment['id'])

        self.assertEqual(restored_comment['status'], 'active')
        self.assertEqual(len(restored_comment['versions']), 2)

    @run_as_sync
    async def test_sample_discussion_related_comments(self):
        """Test getting related comments"""

        family_id = await self.flayer.create_family(
            external_ids={PRIMARY_EXTERNAL_ORG: 'f_external_id'},
        )
        # This will create participant, sample, assay, sequencing group
        participant = await self.player.upsert_participant(get_participant_to_insert())

        sample = participant.samples[0]
        sequencing_group = sample.sequencing_groups[0]
        assay = sequencing_group.assays[0]
        project_id = self.connection.project_id
        assert project_id
        project_name = self.connection.project_id_map[project_id].name

        pat_pid = (
            await self.player.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01_pat'}, reported_sex=1
                )
            )
        ).id

        mat_pid = (
            await self.player.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01_mat'}, reported_sex=2
                )
            )
        ).id

        await self.player.add_participant_to_family(
            family_id=family_id,
            participant_id=participant.id,
            paternal_id=pat_pid,
            maternal_id=mat_pid,
            affected=2,
        )

        sample_comment = await self.add_comment_to_sample(
            sample.to_external().id, 'Sample Comment'
        )

        assay_comment = await self.add_comment_to_assay(assay.id, 'Assay Comment')

        family_comment = await self.add_comment_to_family(family_id, 'Family Comment')

        participant_comment = await self.add_comment_to_participant(
            participant.id, 'Participant Comment'
        )

        project_comment = await self.add_comment_to_project(
            self.connection.project_id, 'Project Comment'
        )

        sequencing_group_comment = await self.add_comment_to_sequencing_group(
            sequencing_group.to_external().id, 'Sg Comment'
        )

        all_comments = [
            (sample_comment, 'Sample'),
            (assay_comment, 'Assay'),
            (family_comment, 'Family'),
            (participant_comment, 'Participant'),
            (project_comment, 'Project'),
            (sequencing_group_comment, 'Sequencing Group'),
        ]

        # add some threaded comments to test those
        for comment, entity in all_comments:
            await self.add_comment_to_thread(
                parent_id=comment['id'], content=f'{entity} threaded comment'
            )

        sample_discussion = await self.get_discussion_on_sample(sample.to_external().id)
        assay_discussion = await self.get_discussion_on_assay(assay.id)
        family_discussion = await self.get_discussion_on_family(family_id)
        participant_discussion = await self.get_discussion_on_participant(
            participant.id
        )
        project_discussion = await self.get_discussion_on_project(project_name)
        sequencing_group_discussion = await self.get_discussion_on_sequencing_group(
            sequencing_group.to_external().id
        )

        # Sample relatedness
        sample_related_ids = set(
            comment_uniq(d) for d in sample_discussion['relatedComments']
        )
        self.assertEqual(len(sample_discussion['directComments']), 1)
        self.assertEqual(len(sample_discussion['relatedComments']), 4)
        self.assertIn(comment_uniq(family_comment), sample_related_ids)
        self.assertIn(comment_uniq(assay_comment), sample_related_ids)
        self.assertIn(comment_uniq(participant_comment), sample_related_ids)
        self.assertIn(comment_uniq(sequencing_group_comment), sample_related_ids)
        self.assertNotIn(comment_uniq(project_comment), sample_related_ids)

        # assay relatedness
        assay_related_ids = set(
            comment_uniq(d) for d in assay_discussion['relatedComments']
        )
        self.assertEqual(len(assay_discussion['directComments']), 1)
        self.assertEqual(len(assay_discussion['relatedComments']), 4)
        self.assertIn(comment_uniq(family_comment), assay_related_ids)
        self.assertIn(comment_uniq(sample_comment), assay_related_ids)
        self.assertIn(comment_uniq(participant_comment), assay_related_ids)
        self.assertIn(comment_uniq(sequencing_group_comment), assay_related_ids)
        self.assertNotIn(comment_uniq(project_comment), assay_related_ids)

        # family relatedness
        family_related_ids = set(
            comment_uniq(d) for d in family_discussion['relatedComments']
        )
        self.assertEqual(len(family_discussion['directComments']), 1)
        self.assertEqual(len(family_discussion['relatedComments']), 4)
        self.assertIn(comment_uniq(assay_comment), family_related_ids)
        self.assertIn(comment_uniq(sample_comment), family_related_ids)
        self.assertIn(comment_uniq(participant_comment), family_related_ids)
        self.assertIn(comment_uniq(sequencing_group_comment), family_related_ids)
        self.assertNotIn(comment_uniq(project_comment), family_related_ids)

        # participant relatedness
        participant_related_ids = set(
            comment_uniq(d) for d in participant_discussion['relatedComments']
        )
        self.assertEqual(len(participant_discussion['directComments']), 1)
        self.assertEqual(len(participant_discussion['relatedComments']), 4)
        self.assertIn(comment_uniq(assay_comment), participant_related_ids)
        self.assertIn(comment_uniq(sample_comment), participant_related_ids)
        self.assertIn(comment_uniq(family_comment), participant_related_ids)
        self.assertIn(comment_uniq(sequencing_group_comment), participant_related_ids)
        self.assertNotIn(comment_uniq(project_comment), participant_related_ids)

        # project relatedness
        project_related_ids = set(
            comment_uniq(d) for d in project_discussion['relatedComments']
        )
        self.assertEqual(len(project_discussion['directComments']), 1)
        self.assertEqual(len(project_discussion['relatedComments']), 5)
        self.assertIn(comment_uniq(assay_comment), project_related_ids)
        self.assertIn(comment_uniq(participant_comment), project_related_ids)
        self.assertIn(comment_uniq(sample_comment), project_related_ids)
        self.assertIn(comment_uniq(family_comment), project_related_ids)
        self.assertIn(comment_uniq(sequencing_group_comment), project_related_ids)

        # sg relatedness
        sequencing_group_related_ids = set(
            comment_uniq(d) for d in sequencing_group_discussion['relatedComments']
        )
        self.assertEqual(len(sequencing_group_discussion['directComments']), 1)
        self.assertEqual(len(sequencing_group_discussion['relatedComments']), 4)
        self.assertIn(comment_uniq(assay_comment), sequencing_group_related_ids)
        self.assertIn(comment_uniq(sample_comment), sequencing_group_related_ids)
        self.assertIn(comment_uniq(family_comment), sequencing_group_related_ids)
        self.assertIn(comment_uniq(participant_comment), sequencing_group_related_ids)
        self.assertNotIn(comment_uniq(project_comment), sequencing_group_related_ids)

    @run_as_sync
    async def test_project_deletion(self):
        """Test ProjectPermissionsTable.delete_project_data's effect on comments"""

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        assay = await self.assaylayer.upsert_assay(
            assay=AssayUpsertInternal(
                sample_id=sample.id, meta=default_sequencing_meta, type='sequencing'
            )
        )
        assert assay.id
        await self.add_comment_to_assay(assay.id, 'Assay comment')

        family = await self.flayer.create_family(
            external_ids={PRIMARY_EXTERNAL_ORG: 'f_external_id'},
        )
        await self.add_comment_to_family(family, 'Family comment')

        participant = await self.player.upsert_participant(
            participant=ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'p_external_id'},
                meta={'pmeta': 'pvalue'},
                reported_sex=2,
                reported_gender='FEMALE',
                karyotype='XX',
            )
        )
        assert participant.id
        await self.add_comment_to_participant(participant.id, 'Participant comment')

        project = self.connection.all_projects()[0]
        await self.add_comment_to_project(project.id, 'Project comment')

        sample_external = sample.to_external()
        await self.add_comment_to_sample(sample_external.id, 'Sample comment')

        sequencing_groups = await self.sglayer.upsert_sequencing_groups(
            sequencing_groups=[
                SequencingGroupUpsertInternal(
                    sample_id=sample.id,
                    type='mtseq',
                    technology='short-read',
                    platform='oxford-nanopore',
                    assays=[
                        AssayUpsertInternal(
                            sample_id=sample.id,
                            meta=default_sequencing_meta,
                            type='sequencing',
                        )
                    ],
                )
            ]
        )
        sequencing_group = sequencing_groups[0]
        sequencing_group_id = sequencing_group.to_external().id
        assert sequencing_group_id
        await self.add_comment_to_sequencing_group(sequencing_group_id, 'SG comment')

        self.assertEqual(await self.row_count('comment'), 6)
        self.assertEqual(await self.row_count('assay_comment'), 1)
        self.assertEqual(await self.row_count('family_comment'), 1)
        self.assertEqual(await self.row_count('participant_comment'), 1)
        self.assertEqual(await self.row_count('project_comment'), 1)
        self.assertEqual(await self.row_count('sample_comment'), 1)
        self.assertEqual(await self.row_count('sequencing_group_comment'), 1)

        pttable = ProjectPermissionsTable(self.connection)
        self.assertTrue(await pttable.delete_project_data(project))

        self.assertEqual(await self.row_count('comment'), 0)
        self.assertEqual(await self.row_count('assay_comment'), 0)
        self.assertEqual(await self.row_count('family_comment'), 0)
        self.assertEqual(await self.row_count('participant_comment'), 0)
        self.assertEqual(await self.row_count('project_comment'), 0)
        self.assertEqual(await self.row_count('sample_comment'), 0)
        self.assertEqual(await self.row_count('sequencing_group_comment'), 0)
